#!/usr/bin/env bash
#
# Bash BASIC-to-C converter --= BACON =--
#
# Peter van Eerten - March/November 2009. License: GPL version 2.
#
# Global variables should be preceded with a 'g_'.
#
# Make sure to edit this file with UTF-8 switched OFF.
# For VIM: set fileencodings=latin1
#
#-----------------------------------------------------------
# CREDITS:
#   - James C. Fuller for endless testing and patiently pointing
#	to many issues
#   - John Spikowksi for providing the forum and giving some hints
#	to improve the language
#
#-----------------------------------------------------------
# GLOBAL INITIALIZATIONS
#-----------------------------------------------------------

# Set the extending globbing option in BASH
shopt -s extglob

# Make sure we can use HIGH ASCII in MacOSX/Ubuntu
export LC_CTYPE="C"
export LANG="C"

# Version of BACON
g_VERSION="0.110"

# Find 'grep' and other coretools
if [[ -z `which grep` || -z `which cat` || -z `which rm` || -z `which tr` || -z `which touch` || -z `which uname` ]]
then
    echo "ERROR: 'grep', 'cat', 'rm', 'tr', 'touch' or 'uname' not found on this system!"
    exit 1
fi

# Link flags
if [[ `uname` = "OSF1" || `uname` = +(*BSD*) ]]
then
    g_LDFLAGS="-lm"
else
    g_LDFLAGS="-lm -ldl"
fi

#-----------------------------------------------------------

function Print_Element
{
    local DQ FUNC

    # Get the element without (surrounding) spaces
    VAR=`echo "${1}" | tr -d "\040"`
    DQ="${VAR: -1}"
    FUNC="${VAR:0:3}"

    # Check type of var, func using string but returning a value?
    if [[ ${FUNC} = "INS" || ${FUNC} = "VAL" || ${FUNC} = "LEN" || ${FUNC} = "SEA" || ${FUNC} = "REG" || ${FUNC} = "ASC" || "$1" = +(FILE*) ]]
    then
	echo "fprintf(stdout, \"%.10g\", (double)($1));" >> $g_CFILE
    # String function?
    elif [[ "$1" = +(*\$*) && "$1" = +(*\(*\)*) ]]
    then
	echo "fprintf(stdout, \"%s\", $1);" >> $g_CFILE
    # Is there a "$" symbol? Then it is always a string, or stringvariable
    elif [[ "$1" = +(*\$*) ]]
    then
	echo "if($1 != NULL) fprintf(stdout, \"%s\", $1);" >> $g_CFILE
	echo "else fprintf(stdout, \"\");" >> $g_CFILE
    # Check if it is a normal string
    elif [[ $DQ = "\"" ]]
    then
	echo "fprintf(stdout, \"%s\", $1);" >> $g_CFILE
    # Other numeric variable
    else
	echo "fprintf(stdout, \"%.10g\", (double)($1));" >> $g_CFILE
    fi
}

#-----------------------------------------------------------

function Handle_Print
{
    local FORMAT EXP LEN TOKEN LINE IN_STRING IN_FUNC CHAR ESCAPED

    IN_FUNC=0

    # If no argument, do nothing
    if [[ "$1" != "PRINT" ]]
    then

	# Get expression without ;
	if [[ -z ${1##*;} ]]
	then
	    let LEN="${#1}"-1
	    EXP="${1:0:$LEN}"
	else
	    EXP="${1}"
	fi

	# If there is FORMAT/format argument
	if [[ "$EXP" = +(* FORMAT *) ]]
	then
	    FORMAT=${EXP##* FORMAT}
	    echo "fprintf(stdout, ${FORMAT%%;*}, ${EXP%%FORMAT *});" >> $g_CFILE
	else
	    # Start miniparser, convert spaces
	    LINE=`echo "${EXP}" | tr " " ""`
	    TOKEN=
	    LEN=${#LINE}

	    # Get the characters
	    until [[ $LEN -eq 0 ]]
	    do
		CHAR="${LINE:0:1}"
		case $CHAR in
		    ","|"+")
			if [[ $IN_STRING -eq 0 && $IN_FUNC -eq 0 ]]
			then
			    Print_Element "${TOKEN}"
			    TOKEN=
			    CHAR=
			    ESCAPED=0
			fi;;
		    "\\")
			ESCAPED=1;;
		    "\"")
			if [[ $ESCAPED -eq 0 ]]
			then
			    if [[ $IN_STRING -eq 0 ]]
			    then
				IN_STRING=1
			    else
				IN_STRING=0
			    fi
			fi
			ESCAPED=0;;
		    "(")
			if [[ $IN_STRING -eq 0 ]]
			then
			    ((IN_FUNC=$IN_FUNC+1))
			fi
			ESCAPED=0;;
		    ")")
			if [[ $IN_STRING -eq 0 ]]
			then
			    ((IN_FUNC=$IN_FUNC-1))
			fi
			ESCAPED=0;;
		    *)
			ESCAPED=0;;
		esac
		# Convert back to space
		if [[ $CHAR = "" ]]
		then
		    TOKEN="${TOKEN} "
		else
		    TOKEN="${TOKEN}${CHAR}"
		fi
		let LEN=${#LINE}-1
		LINE="${LINE: -$LEN}"
	    done
	    Print_Element "${TOKEN}"

	    # If line ends with ';' then skip newline
	    if [[ -n ${1##*;} ]]
	    then
		echo "fprintf(stdout, \"\n\");" >> $g_CFILE
	    fi
	fi
    else
	echo "fprintf(stdout, \"\n\");" >> $g_CFILE
    fi

    # Flush buffer
    echo "fflush(stdout);" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Input
{
    # Local variables
    local CHECK VAR TARGET STR LINE LEN IN_STRING IN_FUNC CHAR ESCAPED

    IN_FUNC=0

    # Check if we have an argument at all
    if [[ "$1" = "INPUT" ]]
    then
	echo -e "\nERROR: empty INPUT at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Start miniparser to see if we need to print something, convert spaces
    LINE=`echo "${1}" | tr " " ""`
    VAR=
    LEN=${#LINE}

    # Get the characters
    until [[ $LEN -eq 0 ]]
    do
	CHAR="${LINE:0:1}"
	case $CHAR in
	    ","|"+")
		if [[ $IN_STRING -eq 0 && $IN_FUNC -eq 0 ]]
		then
		    Print_Element "${VAR}"
		    VAR=
		    CHAR=
		    ESCAPED=0
		fi;;
	    "\\")
		ESCAPED=1;;
	    "\"")
		if [[ $ESCAPED -eq 0 ]]
		then
		    if [[ $IN_STRING -eq 0 ]]
		    then
			IN_STRING=1
		    else
			IN_STRING=0
		    fi
		fi
		ESCAPED=0;;
	    "(")
		if [[ $IN_STRING -eq 0 ]]
		then
		    ((IN_FUNC=$IN_FUNC+1))
		fi
		ESCAPED=0;;
	    ")")
		if [[ $IN_STRING -eq 0 ]]
		then
		    ((IN_FUNC=$IN_FUNC-1))
		fi
		ESCAPED=0;;
	    *)
		ESCAPED=0;;
	esac
	# Convert back to space
	if [[ $CHAR = "" ]]
	then
	    VAR="${VAR} "
	else
	    VAR="${VAR}${CHAR}"
	fi
	let LEN=${#LINE}-1
	LINE="${LINE: -$LEN}"
    done

    # Check type of var, string?
    if [[ "${VAR}" = +(*\$) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$VAR = NULL;" >> $g_HFILE
	fi
	# Translate function to C function
	echo "fgets(__b2c__input__buffer, $g_BUFFER_SIZE, stdin);" >> $g_CFILE
	# Always realloc VAR to correct size, maybe it was resized somewhere else
	echo "$VAR = (char*)realloc($VAR, $g_BUFFER_SIZE*sizeof(char));" >> $g_CFILE
	# Make sure internal var is copied to var of program
	echo "strncpy($VAR, __b2c__input__buffer, $g_BUFFER_SIZE);" >> $g_CFILE
	# Get rid of the terminating newline
	echo "$VAR[strlen(__b2c__input__buffer)-1] = '\0';" >> $g_CFILE
    # Var is numeric or stringarray
    else
	# Variable may not be array, these should be defined with DECLARE
	if [[ "$VAR" != +(*\[*\]*) && "$VAR" != +(*.*) ]]
	then
	    # Not declared? Assume long
	    CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $VAR,| $VAR;|,$VAR,|,$VAR;" | grep -v "\/\*record\*\/"`
	    if [[ -z $CHECK ]]
	    then
		echo "long $VAR;" >> $g_HFILE
		CHECK="long "
	    fi
	else
	    CHECK=`grep -E "DIR |FILE |int |long |float |double |char |void " $TARGET | grep -E " ${VAR%%\[*}" | grep -v "\/\*record\*\/"`
	fi

	# Translate function to C function
	echo "fgets(__b2c__input__buffer, $g_BUFFER_SIZE, stdin);" >> $g_CFILE

	# Make sure internal var is copied to var of program
	if [[ "$CHECK" = +(double *) || "$CHECK" = +(float *) ]]
	then
	    echo "$VAR = atof(__b2c__input__buffer);" >> $g_CFILE
	elif [[ "$CHECK" = +(long *) ]]
	then
	    echo "$VAR = atol(__b2c__input__buffer);" >> $g_CFILE
	elif [[ "$CHECK" = +(int *) ]]
	then
	    echo "$VAR = atoi(__b2c__input__buffer);" >> $g_CFILE
	else
	    echo "$VAR = (char*)realloc($VAR, $g_BUFFER_SIZE*sizeof(char));" >> $g_CFILE
	    echo "strncpy($VAR, __b2c__input__buffer, $g_BUFFER_SIZE);" >> $g_CFILE
	    echo "$VAR[strlen(__b2c__input__buffer)-1] = '\0';" >> $g_CFILE
	fi
    fi
}

#-----------------------------------------------------------

function Handle_For
{
    # Local variables
    local FROM TO TMP VAR STEP CHECK TARGET

    # Get the variablename without (surrounding) spaces
    VAR=`echo "${1%%=*}" | tr -d "\040"`; TMP=`echo "${1#*=}"`

    # Do we have a STRING var?
    if [[ "${VAR}" = +(*\$*) ]]
    then
	echo -e "\nERROR: variable in FOR statement at line $g_COUNTER in file $g_CURFILE cannot be string!"
	exit 1
    fi

    # Check if TO is available
    if [[ "$TMP" != +(* TO *) ]]
    then
	echo -e "\nERROR: missing TO in FOR statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the starting and ending value
    FROM=`echo "${TMP%% TO *}"`
    TO=`echo "${TMP##* TO }"`

    # Check if there is a STEP
    if [[ "$TO" = +(* STEP *) ]]
    then
	STEP="${TO##* STEP }"
	TO="${TO%% STEP *}"
    else
	STEP=1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Variable may not be array, these should be defined with DECLARE
    if [[ "$VAR" != +(*\[*\]*) && "$VAR" != +(*.*) ]]
    then
	# Declare variable if not done yet, assuming long
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $VAR,| $VAR;|,$VAR,|,$VAR;| ${VAR%%\[*}\[|,${VAR%%\[*}\[" | grep -v "\/\*record\*\/"`
	if [[ -z $CHECK ]]
	then
	    echo "long $VAR;" >> $g_HFILE
	fi
    fi

    # Translate function to C function
    if [[ "${STEP}" = +(*-*) ]]
    then
	echo "for($VAR=$FROM; $VAR >= $TO; $VAR+=$STEP){" >> $g_CFILE
    else
	echo "for($VAR=$FROM; $VAR <= $TO; $VAR+=$STEP){" >> $g_CFILE
    fi
}

#-----------------------------------------------------------

function Handle_If
{
    # Check if THEN is available
    if [[ "$1" != +(* THEN) ]]
    then
	echo -e "\nERROR: Missing THEN in IF statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Translate function to C function
    echo "if(${1% *}){" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Elif
{
    # Check if THEN is available
    if [[ "$1" != +(* THEN) ]]
    then
	echo -e "\nERROR: Missing THEN in ELIF statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Translate function to C function
    echo "} else if(${1% *}){" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_While
{
    # Check if DO is available
    if [[ "$1" != +(* DO) ]]
    then
	echo -e "\nERROR: Missing DO in WHILE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    echo "while(${1% *}){" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Let
{
    # Local variables
    local VAR CHECK TMP TARGET LEN STR

    # Check if there is an asignment at all, if not exit
    if [[ "$1" != +(*=*) ]]
    then
	echo -e "\nERROR: could not parse line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%%=*}`; TMP=${1#*=}

    # Check if var is string var, exclude RECORD elements
    if [[ "${VAR}" = +(*\$) && "$VAR" != +(*.*) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$VAR = NULL;" >> $g_HFILE
	fi
    # Assume number, exclude RECORD elements
    elif [[ "$VAR" != +(*\[*\]*) && "$VAR" != +(*.*) && "$VAR" != "ERROR" ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $VAR,| $VAR;|,$VAR,|,$VAR;| ${VAR%%\[*}\[|,${VAR%%\[*}\[" | grep -v "\/\*record\*\/"`
	if [[ -z $CHECK ]]
	then
	    if [[ "$TMP" = +(*.*) ]]
	    then
		echo "double $VAR;" >> $g_HFILE
	    else
		echo "long $VAR;" >> $g_HFILE
	    fi
	fi
    fi

    # Do we have a STRING variable?
    if [[ "${VAR}" = +(*\$*) ]]
    then
	echo "__b2c__assign = (char*) strdup ($TMP); ${g_WITHVAR}${VAR} = (char*)realloc(${g_WITHVAR}${VAR}, (strlen(__b2c__assign)+1)*sizeof(char));" >> $g_CFILE
	echo "strcpy(${g_WITHVAR}${VAR}, __b2c__assign); free(__b2c__assign);" >> $g_CFILE
    else
	echo "${g_WITHVAR}${1};" >> $g_CFILE
    fi
}

#-----------------------------------------------------------

function Handle_Open
{
    # Local variables
    local FILE MODE HANDLE TMP CHECK TARGET LABEL

    # Check if FOR is available
    if [[ "$1" != +(* FOR *) ]]
    then
	echo -e "\nERROR: Missing FOR in OPEN statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if AS is available
    if [[ "$1" != +(* AS *) ]]
    then
	echo -e "\nERROR: Missing AS in OPEN statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the file, mode and handle
    FILE=`echo "${1%% FOR *}"`
    TMP=`echo "${1##* FOR }"`

    MODE=`echo "${TMP%% AS *}"`
    HANDLE=`echo "${TMP##* AS }"`

    # Check if var is string var
    if [[ "${HANDLE}" = +(*\$) ]]
    then
	echo -e "\nERROR: Variable for OPEN at line $g_COUNTER in file $g_CURFILE cannot be string!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check if variable was declared
    if [[ "$HANDLE" != +(*.*) ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $HANDLE,| $HANDLE;|,$HANDLE,|,$HANDLE;| ${HANDLE%%\[*}\[|,${HANDLE%%\[*}\[" | grep -v "\/\*record\*\/"`
    fi

    # File or dir?
    if [[ $MODE = "DIRECTORY" ]]
    then
	if [[ -z $CHECK ]]
	then
	    echo "DIR* $HANDLE;" >> $g_HFILE
	fi
    elif [[ $MODE = "MEMORY" ]]
    then
	if [[ -z $CHECK ]]
	then
	    echo "char* $HANDLE;" >> $g_HFILE
	fi
    elif [[ $MODE = "NETWORK" || $MODE = "SERVER" ]]
    then
	if [[ -z $CHECK ]]
	then
	    echo "int $HANDLE;" >> $g_HFILE
	fi
    else
	if [[ -z $CHECK ]]
	then
	    echo "FILE* $HANDLE;" >> $g_HFILE
	fi
    fi

    # Convert to C syntax
    case $MODE in
	@(READING) )
	    echo "$HANDLE = fopen($FILE, \"r\");" >> $g_CFILE
	    echo "if($HANDLE == NULL){if(!__b2c__trap){ERROR = 2; if(!__b2c__catch_set) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(WRITING) )
	    echo "$HANDLE = fopen($FILE, \"w\");" >> $g_CFILE
	    echo "if($HANDLE == NULL){if(!__b2c__trap){ERROR = 2; if(!__b2c__catch_set) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(APPENDING) )
	    echo "$HANDLE = fopen($FILE, \"a\");" >> $g_CFILE
	    echo "if($HANDLE == NULL){if(!__b2c__trap){ERROR = 2; if(!__b2c__catch_set) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(READWRITE) )
	    echo "$HANDLE = fopen($FILE, \"r+\");" >> $g_CFILE
	    echo "if($HANDLE == NULL){if(!__b2c__trap){ERROR = 2; if(!__b2c__catch_set) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(DIRECTORY) )
	    echo "$HANDLE = opendir($FILE);" >> $g_CFILE
	    echo "if($HANDLE == NULL){if(!__b2c__trap){ERROR = 2; if(!__b2c__catch_set) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(MEMORY) )
	    echo "$HANDLE = (char*)$FILE; long __b2c_mem_$HANDLE; __b2c_mem_$HANDLE = $FILE;" >> $g_CFILE
	    echo "if(!__b2c__trap){__b2c__memory__check($HANDLE); if(ERROR) {if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(NETWORK) )
	    # Network code
	    echo "if (strstr($FILE, \":\") == NULL) {if(!__b2c__trap){ERROR = 10;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "char __b2c__data[$g_BUFFER_SIZE]; strncpy(__b2c__data, $FILE, $g_BUFFER_SIZE);" >> $g_CFILE
	    echo "char *__b2c__host; char *__b2c__port; __b2c__host = strtok(__b2c__data, \":\"); __b2c__port = strtok(NULL, \":\");" >> $g_CFILE
	    echo "struct hostent *__b2c__he; __b2c__he = gethostbyname(__b2c__host);" >> $g_CFILE
	    echo "if (__b2c__he == NULL){if(!__b2c__trap){ERROR = 11;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "$HANDLE = socket(PF_INET, SOCK_STREAM, 0);" >> $g_CFILE
	    echo "if ($HANDLE == -1) {if(!__b2c__trap){ERROR = 12;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "int __b2c__yes = 1; setsockopt($HANDLE, SOL_SOCKET, SO_REUSEADDR, &__b2c__yes, sizeof(int));" >> $g_CFILE
	    echo "struct sockaddr_in __b2c__addr; __b2c__addr.sin_family = AF_INET; __b2c__addr.sin_port = htons((long)atol(__b2c__port));" >> $g_CFILE
	    echo "__b2c__addr.sin_addr = *((struct in_addr *)__b2c__he->h_addr);" >> $g_CFILE
	    echo "memset(&(__b2c__addr.sin_zero), '\0', sizeof(__b2c__addr.sin_zero));" >> $g_CFILE
	    echo "int __b2c__result; __b2c__result = connect($HANDLE, (struct sockaddr *)&__b2c__addr, sizeof(struct sockaddr));" >> $g_CFILE
	    echo "if(__b2c__result == -1) {if(!__b2c__trap){ERROR = 13;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE;;
	@(SERVER) )
	    # Network code
	    echo "if (strstr($FILE, \":\") == NULL){if(!__b2c__trap){ERROR = 10;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "char __b2c__data[$g_BUFFER_SIZE]; strncpy(__b2c__data, $FILE, $g_BUFFER_SIZE);" >> $g_CFILE
	    echo "char *__b2c__host; char *__b2c__port; __b2c__host = strtok(__b2c__data, \":\"); __b2c__port = strtok(NULL, \":\");" >> $g_CFILE
	    echo "struct hostent *__b2c__he; __b2c__he = gethostbyname(__b2c__host);" >> $g_CFILE
	    echo "if (__b2c__he == NULL) {if(!__b2c__trap){ERROR = 11;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "$HANDLE = socket(AF_INET, SOCK_STREAM, 0);" >> $g_CFILE
	    echo "if ($HANDLE == -1) {if(!__b2c__trap){ERROR = 12;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "int __b2c__yes = 1; setsockopt($HANDLE, SOL_SOCKET, SO_REUSEADDR, &__b2c__yes, sizeof(int));" >> $g_CFILE
	    echo "struct sockaddr_in __b2c__addr; __b2c__addr.sin_family = AF_INET; __b2c__addr.sin_port = htons((long)atol(__b2c__port));" >> $g_CFILE
	    echo "__b2c__addr.sin_addr = *((struct in_addr *)__b2c__he->h_addr);" >> $g_CFILE
	    echo "memset(&(__b2c__addr.sin_zero), '\0', sizeof(__b2c__addr.sin_zero));" >> $g_CFILE
	    echo "int __b2c__result; __b2c__result = bind($HANDLE, (struct sockaddr *)&__b2c__addr, sizeof(struct sockaddr));" >> $g_CFILE
	    echo "if(__b2c__result == -1){if(!__b2c__trap){ERROR = 17;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "__b2c__result = listen($HANDLE, $g_MAX_BACKLOG);" >> $g_CFILE
	    echo "if(__b2c__result == -1){if(!__b2c__trap){ERROR = 18;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "__b2c__result = accept($HANDLE, NULL, 0);" >> $g_CFILE
	    echo "if(__b2c__result == -1){if(!__b2c__trap){ERROR = 19;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	    echo "$HANDLE= __b2c__result;" >> $g_CFILE;;
    esac
}

#-----------------------------------------------------------

function Handle_Readln
{
    # Local variables
    local CHECK VAR FROM TARGET STR

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: Missing FROM in READLN statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% FROM *}`

    # Get filedescriptor
    FROM=`echo ${1##* FROM }`

    # Check if var is string var
    if [[ "${VAR}" != +(*\$*) ]]
    then
	echo -e "\nERROR: Variable for READLN at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check if variable is declared
    if [[ ${VAR} != +(*\[*\]*) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$VAR = NULL;" >> $g_HFILE
	fi
    fi

    # Translate function to C function
    echo "fgets(__b2c__input__buffer, $g_BUFFER_SIZE, $FROM);" >> $g_CFILE

    # Make sure internal var is copied to var of program
    echo "$VAR = (char*)realloc($VAR, $g_BUFFER_SIZE*sizeof(char));" >> $g_CFILE
    echo "strncpy($VAR, __b2c__input__buffer, $g_BUFFER_SIZE);" >> $g_CFILE
    # Make sure to end the string
    echo "$VAR[strlen(__b2c__input__buffer)-1] = '\0';" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Writeln
{
    # Local variables
    local VAR TO LINE LEN CHAR IN_STRING IN_FUNC

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in WRITELN statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get filedescriptor
    TO=`echo ${1##* TO }`

    # Start miniparser to see if we need to print something, convert spaces
    LINE=`echo "${1%% TO *}" | tr " " ""`
    VAR=
    LEN=${#LINE}
    IN_STRING=0

    # Get the characters
    until [[ $LEN -eq 0 ]]
    do
	CHAR="${LINE:0:1}"
	case $CHAR in
	    ","|"+")
		if [[ $IN_STRING -eq 0 && $IN_FUNC -eq 0 ]]
		then
		    # Check if var is string var
		    if [[ "${VAR}" = +(*\$*) || "${VAR}" != `echo "${VAR}" | tr -d "\042"` ]]
		    then
			echo "fprintf($TO, \"%s\", ${VAR});" >> $g_CFILE
			VAR=
			CHAR=
			ESCAPED=0
		    else
			echo -e "\nERROR: Argument for WRITELN at line $g_COUNTER in file $g_CURFILE must be string!"
			exit 1
		    fi
		fi;;
	    "\\")
		ESCAPED=1;;
	    "\"")
		if [[ $ESCAPED -eq 0 ]]
		then
		    if [[ $IN_STRING -eq 0 ]]
		    then
			IN_STRING=1
		    else
			IN_STRING=0
		    fi
		fi
		ESCAPED=0;;
	    "(")
		if [[ $IN_STRING -eq 0 ]]
		then
		    ((IN_FUNC=$IN_FUNC+1))
		fi
		ESCAPED=0;;
	    ")")
		if [[ $IN_STRING -eq 0 ]]
		then
		    ((IN_FUNC=$IN_FUNC-1))
		fi
		ESCAPED=0;;
	    *)
		ESCAPED=0;;
	esac
	# Convert back to space
	if [[ ${CHAR} = "" ]]
	then
	    VAR="${VAR} "
	else
	    VAR="${VAR}${CHAR}"
	fi
	let LEN=${#LINE}-1
	LINE="${LINE: -$LEN}"
    done

    # Check if var is string var
    if [[ "${VAR}" = +(*\$*) || "${VAR}" != `echo "${VAR}" | tr -d "\042"` ]]
    then
	echo "fprintf($TO, \"%s\n\", ${VAR});" >> $g_CFILE
	echo "fflush($TO);" >> $g_CFILE
    else
	echo -e "\nERROR: Argument for WRITELN at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi
}

#-----------------------------------------------------------

function Handle_Getbyte
{
    # Local variables
    local POS FROM SIZE TARGET CHECK

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: Missing FROM in GETBYTE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    POS=`echo ${1%% FROM *}`

    # Get filedescriptor
    FROM=`echo ${1##* FROM }`

    SIZE=

    # Check if SIZE is available
    if [[ "$1" = +(* SIZE *) ]]
    then
	SIZE=${FROM##* SIZE }
	FROM=${FROM%% SIZE *}
    else
	SIZE=1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Declare variable if not done yet, assuming long
    if [[ "$POS" != +(*.*) ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $POS,| $POS;|,$POS,|,$POS;| ${POS%%\[*}\[|,${POS%%\[*}\[" | grep -v "\/\*record\*\/"`
	if [[ -z $CHECK ]]
	then
	    echo "long $POS;" >> $g_HFILE
	fi
    fi

    # Translate function to C function
    echo "if(!__b2c__trap){__b2c__memory__check((char *)$POS); if(ERROR) {if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
    echo "fread((void*)$POS, sizeof(char), $SIZE, $FROM);" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Putbyte
{
    # Local variables
    local VAR TO SIZE POS

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in PUTBYTE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    POS=`echo ${1%% TO *}`;

    # Get filedescriptor
    TO=`echo ${1##* TO }`

    SIZE=

    # Check if SIZE is available
    if [[ "$1" = +(* SIZE *) ]]
    then
	SIZE=${TO##* SIZE }
	TO=${TO%% SIZE *}
    else
	SIZE=1
    fi

    # Translate function to C function
    echo "if(!__b2c__trap){__b2c__memory__check((char *)$POS); if(ERROR){if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
    echo "fwrite((void*)$POS, sizeof(char), $SIZE, $TO);" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Getfile
{
    # Local variables
    local VAR FROM SIZE TARGET STR

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: Missing FROM in GETFILE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% FROM *}`

    # Get dirdescriptor
    FROM=`echo ${1##* FROM }`

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Translate function to C function
    echo "struct dirent *__b2c__dir; __b2c__dir = readdir($FROM);" >> $g_CFILE

    # Check if variable is declared
    STR=${VAR%\$*}
    CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
    if [[ -z $CHECK ]]
    then
	echo "char *$VAR = NULL;" >> $g_HFILE
    fi

    # Always realloc VAR to correct size, maybe it was resized somewhere else
    echo "if(__b2c__dir != NULL) {$VAR = realloc($VAR, (strlen(__b2c__dir->d_name)+1)*sizeof(char));" >> $g_CFILE
    # Make sure internal var is copied to var of program
    echo "strcpy($VAR, __b2c__dir->d_name); $VAR[strlen(__b2c__dir->d_name)] = '\0';}" >> $g_CFILE
    echo "else {$VAR[0] = '\0';}" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Receive
{
    # Local variables
    local CHECK VAR FROM TARGET STR

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: Missing FROM in RECEIVE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% FROM *}`

    # Get filedescriptor
    FROM=`echo ${1##* FROM }`

    # Check if var is string var
    if [[ "${VAR}" != +(*\$) ]]
    then
	echo -e "\nERROR: Variable for RECEIVE at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check if variable is declared
    STR=${VAR%\$*}
    CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
    if [[ -z $CHECK ]]
    then
	echo "char *$VAR = NULL;" >> $g_HFILE
    fi

    # Translate function to C function
    echo "$VAR = (char*)realloc($VAR, ($g_BUFFER_SIZE+1)*sizeof(char));" >> $g_CFILE
    echo "memset($VAR, '\0', $g_BUFFER_SIZE+1);" >> $g_CFILE
    echo "if(recv($FROM, $VAR, $g_BUFFER_SIZE, 0) < 0) {if(!__b2c__trap){ERROR = 14;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
}
#-----------------------------------------------------------

function Handle_Send
{
    # Local variables
    local VAR TO

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in SEND statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% TO *}`;

    # Get filedescriptor
    TO=`echo ${1##* TO }`

    # Check if var is string var
    if [[ "${VAR}" != +(*\$*)  && "$VAR" = `echo "$VAR" | tr -d "\042"` ]]
    then
	echo -e "\nERROR: Variable for SEND at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi

    # Translate function to C function
    echo "if(send($TO, $VAR, strlen($VAR), 0) < 0) {if(!__b2c__trap){ERROR = 15; if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Getline
{
    # Local variables
    local CHECK VAR FROM TARGET STR

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: Missing FROM in GETLINE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% FROM *}`

    # Get filedescriptor
    FROM=`echo ${1##* FROM }`

    # Check if var is string var
    if [[ "${VAR}" != +(*\$*) ]]
    then
	echo -e "\nERROR: Variable for GETLINE at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check if variable is declared
    if [[ ${VAR} != +(*\[*\]*) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$VAR = NULL;" >> $g_HFILE
	fi
    fi

    # Translate function to C function
    echo "if ($FROM == NULL || *$FROM == '\0') {$VAR = (char*)realloc($VAR, 2*sizeof(char)); strcpy($VAR, \"\");}" >> $g_CFILE
    echo "else { __b2c__assign = $FROM; while(*$FROM != '\0' && *$FROM != '\n') $FROM++;" >> $g_CFILE

    # Make sure internal var is copied to var of program
    echo "$VAR = (char*)realloc($VAR, ($FROM-__b2c__assign+1)*sizeof(char));" >> $g_CFILE
    echo "strncpy($VAR, __b2c__assign, $FROM-__b2c__assign);" >> $g_CFILE

    # Make sure to end the string
    echo "$VAR[($FROM-__b2c__assign)] = '\0'; $FROM++;}" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Putline
{
    # Local variables
    local VAR TO

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in PUTLINE statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    VAR=`echo ${1%% TO *}`;

    # Get memorydescriptor
    TO=`echo ${1##* TO }`

    # Check if var is string var
    if [[ "${VAR}" != +(*\$*)  && "$VAR" = `echo "$VAR" | tr -d "\042"` ]]
    then
	echo -e "\nERROR: Variable for PUTLINE at line $g_COUNTER in file $g_CURFILE must be string!"
	exit 1
    fi

    # Translate function to C function
    echo "strcat($TO, $VAR); strcat($TO, \"\n\"); $TO+=strlen($VAR)+1;" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Import
{
    # Local variables
    local TMP SYM LIB CHECK TOKEN PTR TYPE

    # Check if FROM is available
    if [[ "$1" != +(* FROM *) ]]
    then
	echo -e "\nERROR: missing FROM in IMPORT statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the symbolname without surrounding spaces and doublequotes
    if [[ "$1" = +(*\(*) ]]
    then
	SYM=`echo ${1%%\(*} | tr -d "\042"`; TMP=${1#* FROM }
	TOKEN=`echo ${1%%\)*}`
    else
	SYM=`echo ${1%% FROM *} | tr -d "\042"`; TMP=${1#* FROM }
	TOKEN=
    fi

    # Check if TYPE is available
    if [[ "$1" != +(* TYPE *) ]]
    then
	echo -e "\nERROR: missing TYPE in IMPORT statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get library and type
    LIB=`echo ${TMP%% TYPE *}`
    TYPE=`echo ${TMP##* TYPE }`

    # If library is libm or libc, skip dlopen as we're linking with those anyway
    if [[ $LIB != +(*libc.so*) && $LIB != +(*libm.so*) ]]
    then
	# Create name from libname
	PTR=`echo $LIB | tr -d [:punct:]`

	# Check if variable was declared
	CHECK=`grep -i "void\* __b2c__dlopen__pointer_$PTR;" $g_HFILE`
	if [[ -z $CHECK ]]
	then
	    echo "void* __b2c__dlopen__pointer_$PTR;" >> $g_HFILE
	    echo "__b2c__dlopen__pointer_$PTR = dlopen($LIB, RTLD_GLOBAL|RTLD_LAZY);" >> $g_CFILE
	    echo "if(__b2c__dlopen__pointer_$PTR == NULL){if(!__b2c__trap){ERROR = 3;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
	fi

	# Check if token was declared
	CHECK=`grep -i "$TYPE (\*$SYM)" $g_HFILE`
	if [[ -z $CHECK ]]
	then
	    echo "$TYPE (*$SYM)(${TOKEN##*\(});" >> $g_HFILE
	fi

	# Translate to C function
	echo "*($TYPE **) (&$SYM) = dlsym(__b2c__dlopen__pointer_$PTR, \"$SYM\");" >> $g_CFILE
	echo "if($SYM == NULL) {if(!__b2c__trap){ERROR = 4;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
    fi

    # Make symbol known to parser
    g_IMPORTED="$SYM $g_IMPORTED"
}

#-----------------------------------------------------------

function Handle_Declare
{
    # Local variables
    local VAR TYPE CHECK

    # Get the variablename and type
    if [[ "$1" = +(* TYPE *) ]]
    then
	VAR=`echo ${1%% TYPE *}`
	TYPE=`echo ${1##* TYPE }`
    else
	VAR=`echo ${1}`
	TYPE="long"
    fi

    # Check if variable was already declared
    if [[ "$VAR" != +(*.*) ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $g_HFILE | grep -E " $VAR,| $VAR;|,$VAR,|,$VAR;| ${VAR%%\[*}\[|,${VAR%%\[*}\[" | grep -v "\/\*record\*\/"`
    fi

    if [[ -n $CHECK ]]
    then
	echo -e "\nERROR: variable in DECLARE or GLOBAL statement at line $g_COUNTER in file $g_CURFILE was defined previously!"
	exit 1
    fi

    # Check if var is string var
    if [[ "${VAR}" = +(*\$*) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" $g_HFILE`
	if [[ -z $CHECK ]]
	then
	    STR=`echo ${VAR} | tr ',' ' '`
	    for i in $STR
	    do
		if [[ "${i}" = +(*\[*\]*) ]]
		then
		    echo "char *$i = { NULL };" >> $g_HFILE
		else
		    echo "char *$i = NULL;" >> $g_HFILE
		fi
	    done
	fi
    # Assume number
    else
	echo "$TYPE $VAR;" >> $g_HFILE
    fi
}

#-----------------------------------------------------------

function Handle_Local
{
    # Local variables
    local VAR TYPE CHECK TARGET DIM

    # Get the variablename and type
    if [[ "$1" = +(* TYPE *) ]]
    then
	VAR=`echo ${1%% TYPE *}`
	TYPE=`echo ${1##* TYPE }`
    else
	VAR=`echo ${1}`
	TYPE="long"
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET=$g_CFILE
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check if variable was already declared
    if [[ "$VAR" != +(*.*) ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $VAR,| $VAR;|,$VAR,|,$VAR;| ${VAR%%\[*}\[|,${VAR%%\[*}\[" | grep -v "\/\*record\*\/"`
    fi

    if [[ -n $CHECK ]]
    then
	echo -e "\nERROR: variable in LOCAL statement at line $g_COUNTER in file $g_CURFILE was defined previously!"
	exit 1
    fi

    # Check if var is string var
    if [[ "${VAR}" = +(*\$*) && "${VAR}" != +(*\[*\$*\]*) ]]
    then
	STR=${VAR%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = " ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    STR=`echo ${VAR} | tr ',' ' '`
	    for i in $STR
	    do
		# Check on multidimensional stringarrays
		if [[ "${i}" = +(*\[*\]*\]*) ]]
		then
		    echo -e "\nERROR: multidimensional stringarrays at line $g_COUNTER in file $g_CURFILE are not supported!"
		    exit 1
		fi
		# Are we in a function?
		if [[ -n $g_FUNCNAME ]]
		then
		    if [[ -n $g_RECORDNAME ]]
		    then
			if [[ "${i}" = +(*\[*\]*) ]]
			then
			    echo "char *$i;" >> $g_CFILE
			    DIM=${i##*\[}; DIM=${DIM%%\]*}
			    g_STRINGARRAYS="$g_STRINGARRAYS for(__b2c__ctr=0; __b2c__ctr<$DIM; __b2c__ctr++)if($g_RECORDVAR.${i%%\[*}[__b2c__ctr]!=NULL){free($g_RECORDVAR.${i%%\[*}[__b2c__ctr]);}"
			else
			    echo "char *$i; /*record*/" >> $g_CFILE
			fi
		    else
			if [[ "${i}" = +(*\[*\]*) ]]
			then
			    echo "char *$i = { NULL };" >> $g_CFILE
			    DIM=${i##*\[}; DIM=${DIM%%\]*}
			    g_STRINGARRAYS="$g_STRINGARRAYS for(__b2c__ctr=0; __b2c__ctr<$DIM; __b2c__ctr++)if(${g_WITHVAR}${i%%\[*}[__b2c__ctr]!=NULL){free(${g_WITHVAR}${i%%\[*}[__b2c__ctr]);}"
			else
			    echo "char *$i = NULL;" >> $g_CFILE
			    g_LOCALSTRINGS="$g_LOCALSTRINGS ${i}"
			fi
		    fi
		# We are in the mainprogram
		else
		    if [[ -n $g_RECORDNAME ]]
		    then
			echo "char *$i; /*record*/" >> $g_HFILE
		    else
			if [[ "${i}" = +(*\[*\]*) ]]
			then
			    echo "char *$i = { NULL };" >> $g_CFILE
			else
			    echo "char *$i = NULL;" >> $g_HFILE
			fi
		    fi
		fi
	    done
	fi
    # Assume number
    else
	if [[ -n $g_FUNCNAME ]]
	then
	    if [[ -n $g_RECORDNAME ]]
	    then
		echo "$TYPE $VAR; /*record*/" >> $g_CFILE
	    else
		echo "$TYPE $VAR;" >> $g_CFILE
	    fi
	else
	    if [[ -n $g_RECORDNAME ]]
	    then
		echo "$TYPE $VAR; /*record*/" >> $g_HFILE
	    else
		echo "$TYPE $VAR;" >> $g_HFILE
	    fi
	fi
    fi
}

#-----------------------------------------------------------

function Handle_Read
{
    # Local variables
    local CHECK TARGET STR

    # Check if we have an argument at all
    if [[ "$1" = "READ" ]]
    then
	echo -e "\nERROR: empty READ at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Check type of var, string?
    if [[ "${1}" = +(*\$) ]]
    then
	STR=${1%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$1 = NULL;" >> $g_HFILE
	fi
	# Convert to C
	echo "$1 = (char*)realloc($1, (strlen(__b2c__stringarray[__b2c__stringarray_ptr])+1)*sizeof(char));" >> $g_CFILE
	echo "strcpy($1, __b2c__stringarray[__b2c__stringarray_ptr]);" >> $g_CFILE
	echo "__b2c__stringarray_ptr++;" >> $g_CFILE
    else
	# Variable may not be array, these should be defined with DECLARE
	if [[ "$1" != +(*\[*\]*) && "$1" != +(*.*) ]]
	then
	    # Not declared? Assume long
	    CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $1,| $1;|,$1,|,$1;| ${1%%\[*}\[|,${1%%\[*}\[" | grep -v "\/\*record\*\/"`
	    if [[ -z $CHECK ]]
	    then
		echo "long $1;" >> $g_HFILE
		CHECK="long "
	    fi
	else
	    # See how var was declared
	    CHECK=`grep -E "DIR |FILE |int |long |float |double |char |void |STRING |NUMBER |FLOATING " $TARGET | grep -E " ${1%%\[*}" | grep -v "\/\*record\*\/"`
	fi
	# Convert to C
	if [[ "$CHECK" = +(double *) || "$CHECK" = +(float *) ]]
	then
	    echo "${1} = __b2c__floatarray[__b2c__floatarray_ptr];" >> $g_CFILE
	    echo "__b2c__floatarray_ptr++;" >> $g_CFILE
	elif [[ "$CHECK" = +(long *) || "$CHECK" = +(int *) ]]
	then
	    echo "${1} = __b2c__intarray[__b2c__intarray_ptr];" >> $g_CFILE
	    echo "__b2c__intarray_ptr++;" >> $g_CFILE
	else
	    echo "${1} = (char*)realloc($1, (strlen(__b2c__stringarray[__b2c__stringarray_ptr])+1)*sizeof(char));" >> $g_CFILE
	    echo "strcpy(${1}, __b2c__stringarray[__b2c__stringarray_ptr]);" >> $g_CFILE
	    echo "__b2c__stringarray_ptr++;" >> $g_CFILE
	fi
    fi
}

#-----------------------------------------------------------

function Handle_Endfunction
{
    # Check if return was found
    if [[ -z $g_FUNCTYPE ]]
    then
	echo -e "\nERROR: function '$g_FUNCNAME' was defined without returning a value or string!"
	exit 1
    fi

    # Now setup function in main program
    if [[ $g_ORIGFUNCNAME != +(*\(*\)*) ]]
    then
	echo "$g_FUNCTYPE $g_ORIGFUNCNAME(void) {" > ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    else
	echo "$g_FUNCTYPE $g_ORIGFUNCNAME {" > ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    fi

    # Add function body
    echo "${g_STRINGARGS}" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    cat $g_CFILE >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    echo "__B2C__PROGRAM__EXIT: ;" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    echo "}" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h

    # Include header file
    g_INCLUDE_FILES="$g_INCLUDE_FILES\n#include \"${g_COPY_CFILE%\.*}.$g_FUNCNAME.h\""

    # Add to total filelist
    g_TMP_FILES="$g_TMP_FILES ${g_COPY_CFILE%.*}.$g_FUNCNAME.h"

    # Delete temp funcfile
    rm $g_CFILE

    # Restore mainfile
    g_CFILE=$g_COPY_CFILE

    # Restore CATCH routine
    g_CATCHGOTO="$g_ORIGCATCHGOTO"

    # Clear function variables
    g_ORIGFUNCNAME=
    g_FUNCNAME=
    g_FUNCTYPE=
    g_LOCALSTRINGS=
    g_STRINGARRAYS=
    g_STRINGARGS=
}

#-----------------------------------------------------------

function Handle_Return
{
    # Local variables
    local CHECK STR DQ FUNC ARG

    # Check if we have an argument at all
    if [[ "$1" = "RETURN" ]]
    then
	echo -e "\nERROR: empty RETURN at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $g_CFILE | grep -E " $1,| $1;|,$1,|,$1;| ${1%%\[*}\[|,${1%%\[*}\[" | grep -v "\/\*record\*\/"`

    # Determine if argument ends with Double Quote
    DQ="${1: -1}"
    FUNC="${1:0:3}"

   # Check type of var, func using string but returning a value? See also PRINT miniparser
    if [[ ${FUNC} = "INS" || ${FUNC} = "VAL" || ${FUNC} = "LEN" || ${FUNC} = "SEA" || ${FUNC} = "ASC" || "${1}" = +(FILE*) || "${1}" = +(REGEX*) ]]
    then
	g_FUNCTYPE="double "
	ARG="${1}"

    # Check type of var, string or normal string?
    elif [[ "${1}" = +(*\$*) || "$DQ" = "\"" || "$DQ" = "\"" ]]
    then
	g_FUNCTYPE="char * "
	echo "__b2c__rbuffer_ptr++; if(__b2c__rbuffer_ptr >= $g_MAX_RBUFFERS) __b2c__rbuffer_ptr=0; if($1 != NULL)" >> $g_CFILE
	echo " {__b2c__rbuffer[__b2c__rbuffer_ptr] = (char*)realloc(__b2c__rbuffer[__b2c__rbuffer_ptr], (strlen($1)+1)*sizeof(char));" >> $g_CFILE
	echo "strcpy(__b2c__rbuffer[__b2c__rbuffer_ptr], $1);}" >> $g_CFILE
	echo "else {__b2c__rbuffer[__b2c__rbuffer_ptr] = (char*)realloc(__b2c__rbuffer[__b2c__rbuffer_ptr], 2*sizeof(char));" >> $g_CFILE
	echo "strcpy(__b2c__rbuffer[__b2c__rbuffer_ptr], \"\");}" >> $g_CFILE
	ARG="__b2c__rbuffer[__b2c__rbuffer_ptr]"

    # Check if float
    elif [[ "$1" = +(*.*) ]]
    then
	g_FUNCTYPE="double "
	ARG="${1}"

    # Check if no alpha chars (then integer value)
    elif [[ "$1" = +([0-9]) ]]
    then
	g_FUNCTYPE="long "
	ARG="${1}"

    # Assume variable, check if declared before
    elif [[ -n $CHECK ]]
    then
	case $CHECK in
	    +(*DIR *) )
		g_FUNCTYPE="DIR* ";;
	    +(*FILE *) )
		g_FUNCTYPE="FILE* ";;
	    +(*int\* *) )
		g_FUNCTYPE="int* ";;
	    +(*float\* *) )
		g_FUNCTYPE="float* ";;
	    +(*double\* *) )
		g_FUNCTYPE="double* ";;
	    +(*char\* *) )
		g_FUNCTYPE="char* ";;
	    +(*long\* *) )
		g_FUNCTYPE="long* ";;
	    +(*void\* *) )
		g_FUNCTYPE="void* ";;
	    +(*int *) )
		g_FUNCTYPE="int ";;
	    +(*float *) )
		g_FUNCTYPE="float ";;
	    +(*double *) )
		g_FUNCTYPE="double ";;
	    +(*char *) )
		g_FUNCTYPE="char ";;
	    +(*long *) )
		g_FUNCTYPE="long ";;
	    +(*void *) )
		g_FUNCTYPE="void ";;
	esac
	ARG="${1}"

    # Not declared, assume integer variable
    else
	g_FUNCTYPE="long "
	ARG="${1}"
    fi

    # Free strings variables if there are any
    for STR in $g_LOCALSTRINGS
    do
	echo "if($STR != NULL) free($STR);" >> $g_CFILE
    done
    echo "${g_STRINGARRAYS}" >> $g_CFILE

    # The actual return value
    echo "return (${ARG});" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Push
{
    # Local variables
    local CHECK TARGET

    # Check if we have an argument at all
    if [[ "$1" = "PUSH" ]]
    then
	echo -e "\nERROR: empty PUSH at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    if [[ "$VAR" != +(*.*) ]]
    then
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $1,| $1;|,$1,|,$1;| ${1%%\[*}\[|,${1%%\[*}\[" | grep -v "\/\*record\*\/"`
    fi

    # Check type of var, string?
    if [[ "${1}" = +(*\$*) ]]
    then
	echo "if(__b2c__stringstack[__b2c__stackptr] == NULL){__b2c__stringstack[__b2c__stackptr] = (char*)realloc(__b2c__stringstack[__b2c__stackptr], (strlen($1)+1)*sizeof(char));}" >> $g_CFILE
	echo "strcpy(__b2c__stringstack[__b2c__stackptr], $1);" >> $g_CFILE
	echo "__b2c__typestack[__b2c__stackptr] = 1;" >> $g_CFILE
    # Check if it is a normal string
    elif [[ "$1" != `echo "$1" | tr -d "\042"` ]]
    then
	echo "if(__b2c__stringstack[__b2c__stackptr] == NULL){__b2c__stringstack[__b2c__stackptr] = (char*)realloc(__b2c__stringstack[__b2c__stackptr], (strlen($1)+1)*sizeof(char));}" >> $g_CFILE
	echo "strcpy(__b2c__stringstack[__b2c__stackptr], $1);" >> $g_CFILE
	echo "__b2c__typestack[__b2c__stackptr] = 1;" >> $g_CFILE
    # Check if float
    elif [[ "$1" = +(*.*) ]]
    then
	echo "__b2c__doublestack[__b2c__stackptr] = $1;" >> $g_CFILE
	echo "__b2c__typestack[__b2c__stackptr] = 2;" >> $g_CFILE
    # Check if no alpha chars (then integer value)
    elif [[ "$1" = +(![a-zA-Z]) ]]
    then
	echo "__b2c__longstack[__b2c__stackptr] = $1;" >> $g_CFILE
	echo "__b2c__typestack[__b2c__stackptr] = 3;" >> $g_CFILE
    # Assume variable, check if declared before
    elif [[ -n $CHECK ]]
    then
	if [[ $CHECK = +(double*) ]]
	then
	    echo "__b2c__doublestack[__b2c__stackptr] = $1;" >> $g_CFILE
	    echo "__b2c__typestack[__b2c__stackptr] = 2;" >> $g_CFILE
	else
	    echo "__b2c__longstack[__b2c__stackptr] = $1;" >> $g_CFILE
	    echo "__b2c__typestack[__b2c__stackptr] = 3;" >> $g_CFILE
	fi
    # Not declared, assume integer variable
    else
	echo "__b2c__longstack[__b2c__stackptr] = $1;" >> $g_CFILE
	echo "__b2c__typestack[__b2c__stackptr] = 3;" >> $g_CFILE
    fi

    # Increase stackpointer
    echo "__b2c__stackptr++;" >> $g_CFILE
    echo "if(__b2c__stackptr == $g_STACK_DEPTH) __b2c__stackptr=$g_STACK_DEPTH-1;" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Pull
{
    local CHECK TARGET STR

    # Check if we have an argument at all
    if [[ "$1" = "PULL" ]]
    then
	echo -e "\nERROR: empty PULL at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Argument must be a variable
    if [[ "$1" = +(![a-zA-Z]) ]]
    then
	echo -e "\nERROR: argument in PULL statement at line $g_COUNTER in file $g_CURFILE is not a variable!"
	exit 1
    fi

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Decrease stackpointer again
    echo "__b2c__stackptr--;" >> $g_CFILE
    echo "if(__b2c__stackptr < 0) __b2c__stackptr=0;" >> $g_CFILE

    # Get the last value from stack
    if [[ "${1}" = +(*\$) ]]
    then
    	STR=${1%\$*}
	CHECK=`grep -E "char \*${STR}\\\\$ = NULL;" ${TARGET}`
	if [[ -z $CHECK ]]
	then
	    echo "char *$1 = NULL;" >> $g_HFILE
	fi
	echo "if(__b2c__typestack[__b2c__stackptr] == 1) {" >> $g_CFILE
	echo "$1 = (char*)realloc($1, (strlen(__b2c__stringstack[__b2c__stackptr])+1)*sizeof(char));" >> $g_CFILE
	echo "strcpy($1, __b2c__stringstack[__b2c__stackptr]);}" >> $g_CFILE
    else
	# Variable may not be array, these should be defined with DECLARE
	if [[ "$1" != +(*\[*\]*) && "$VAR" != +(*.*) ]]
	then
	    # Not declared? Assume long
	    CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $1,| $1;|,$1,|,$1;| ${1%%\[*}\[|,${1%%\[*}\[" | grep -v "\/\*record\*\/"`
	    if [[ -z $CHECK ]]
	    then
		echo "long $1;" >> $g_HFILE
		CHECK="long "
	    fi
	else
	    # See how var was declared
	    CHECK=`grep -E "DIR |FILE |int |long |float |double |char |void |STRING |NUMBER |FLOATING " $TARGET | grep -E " ${1%%\[*}" | grep -v "\/\*record\*\/"`
	fi
	# Make sure internal var is copied to var of program
	if [[ "$CHECK" = +(double *) || "$CHECK" = +(float *) ]]
	then
	    echo "if(__b2c__typestack[__b2c__stackptr] == 2) $1=__b2c__doublestack[__b2c__stackptr];" >> $g_CFILE
	elif [[ "$CHECK" = +(long *) || "$CHECK" = +(int *) ]]
	then
	    echo "if(__b2c__typestack[__b2c__stackptr] == 3) $1=__b2c__longstack[__b2c__stackptr];" >> $g_CFILE
	else
	    echo "if(__b2c__typestack[__b2c__stackptr] == 1) {" >> $g_CFILE
	    echo "$1 = (char*)realloc($1, (strlen(__b2c__stringstack[__b2c__stackptr])+1)*sizeof(char));" >> $g_CFILE
	    echo "strcpy($1, __b2c__stringstack[__b2c__stackptr]);}" >> $g_CFILE
	fi
    fi
}

#-----------------------------------------------------------

function Handle_SubFunc
{
    local ARGS TOKEN LEN CHAR

    # Check argument
    if [[ "${1}" = "SUB" ]]
    then
	echo -e "\nERROR: empty SUB/FUNCTION at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if we are in a sub already
    if [[ -n $g_FUNCNAME ]]
    then
	echo -e "\nERROR: cannot define a SUB/FUNCTION within a SUB/FUNCTION at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the funcname
    g_ORIGFUNCNAME=`echo ${1%%\(*}"("`

    # Start miniparser to duplicate string arguments, convert spaces
    ARGS=`echo "${1##*\(}" | tr " " ""`
    TOKEN=
    LEN=${#ARGS}

    # Get the characters
    until [[ $LEN -eq 0 ]]
    do
	CHAR="${ARGS:0:1}"
	if [[ $CHAR = "," ]]
	then
	    if [[ ${TOKEN} = +(*STRING*) || ${TOKEN} = +(*char\**) ]]
	    then
		g_ORIGFUNCNAME="$g_ORIGFUNCNAME STRING __b2c_${TOKEN##* },"
		g_STRINGARGS="$g_STRINGARGS char* ${TOKEN##* } = strdup(__b2c_${TOKEN##* });"
		g_LOCALSTRINGS="$g_LOCALSTRINGS ${TOKEN##* }"
	    else
		g_ORIGFUNCNAME="$g_ORIGFUNCNAME ${TOKEN},"
	    fi
	    TOKEN=
	    CHAR=
	fi
	# Convert back to space
	if [[ $CHAR = "" ]]
	then
	    TOKEN="${TOKEN} "
	else
	    TOKEN="${TOKEN}${CHAR}"
	fi
	let LEN=${#ARGS}-1
	ARGS="${ARGS: -$LEN}"
    done
    TOKEN=${TOKEN%%\)*}
    if [[ ${TOKEN} = +(*STRING*) || ${TOKEN} = +(*char\**) ]]
    then
	g_ORIGFUNCNAME="$g_ORIGFUNCNAME STRING __b2c_${TOKEN##* })"
	g_STRINGARGS="$g_STRINGARGS char* ${TOKEN##* } = strdup(__b2c_${TOKEN##* });"
	g_LOCALSTRINGS="$g_LOCALSTRINGS ${TOKEN##* }"
    else
	g_ORIGFUNCNAME="$g_ORIGFUNCNAME ${TOKEN})"
    fi

    # Get original function name
    if [[ "$g_ORIGFUNCNAME" = +(* \(*\)*) ]]
    then
	g_FUNCNAME=`echo ${g_ORIGFUNCNAME%% *}`
    elif [[ "$g_ORIGFUNCNAME" = +(*\(*\)*) ]]
    then
	g_FUNCNAME=`echo ${g_ORIGFUNCNAME%%\(*}`
    else
	g_FUNCNAME=`echo $g_ORIGFUNCNAME`
    fi

    # Make symbol known to parser
    g_IMPORTED="$g_FUNCNAME $g_IMPORTED"

    # Switch to header file
    g_COPY_CFILE=$g_CFILE
    g_CFILE=${g_CFILE%.*}.$g_FUNCNAME.tmp

    # Save CATCH routine
    g_ORIGCATCHGOTO="$g_CATCHGOTO"
    g_CATCHGOTO="__B2C__PROGRAM__EXIT"

    touch $g_CFILE
}

#-----------------------------------------------------------

function Handle_Endsub
{
    typeset STR

    # Get original function name
    if [[ "$g_ORIGFUNCNAME" = +(* \(*\)*) ]]
    then
	echo "void ${g_ORIGFUNCNAME} {" > ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    elif [[ "$g_ORIGFUNCNAME" = +(*\(*\)*) ]]
    then
	echo "void ${g_ORIGFUNCNAME} {" > ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    else
	echo "void ${g_FUNCNAME}(void) {" > ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    fi

    # Finalize sub
    echo "${g_STRINGARGS}" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    cat $g_CFILE >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    # Free strings variables if there are any
    for STR in $g_LOCALSTRINGS
    do
	echo "if($STR != NULL) free(${STR});" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    done
    echo "${g_STRINGARRAYS}" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    echo "__B2C__PROGRAM__EXIT: ;" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h
    echo "}" >> ${g_COPY_CFILE%.*}.$g_FUNCNAME.h

    # Include header file
    g_INCLUDE_FILES="$g_INCLUDE_FILES\n#include \"${g_COPY_CFILE%.*}.$g_FUNCNAME.h\""

    # Add to total filelist
    g_TMP_FILES="$g_TMP_FILES ${g_COPY_CFILE%.*}.$g_FUNCNAME.h"

    # Delete temp funcfile
    rm $g_CFILE

    # Restore mainfile
    g_CFILE=$g_COPY_CFILE

    # Restore CATCH routine
    g_CATCHGOTO="$g_ORIGCATCHGOTO"

    # Reset variables
    g_ORIGFUNCNAME=
    g_FUNCNAME=
    g_LOCALSTRINGS=
    g_STRINGARRAYS=
    g_STRINGARGS=
}

#-----------------------------------------------------------

function Handle_Deffn
{
    local SYM
	
    # Check if we have an argument at all
    if [[ "$1" = "DEF" ]]
    then
	echo -e "\nERROR: empty DEF FN at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Translate to C macro
    SYM=${1#* }

    echo "#define ${SYM%%=*} ${SYM#*=}" >> $g_HFILE

    # Make symbol known to parser
    g_IMPORTED="${SYM%%\(*} $g_IMPORTED"
}

#-----------------------------------------------------------

function Handle_Const
{
    # Check if we have an argument at all
    if [[ "$1" = "CONST" ]]
    then
	echo -e "\nERROR: empty CONST at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    echo "#define ${1%%=*} ${1#*=}" >> $g_HFILE
}

#-----------------------------------------------------------

function Handle_Seek
{
    local FILE OFFSET WHENCE

    # Check if we have an argument at all
    if [[ "$1" = "SEEK" ]]
    then
	echo -e "\nERROR: empty SEEK at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if TO is available
    if [[ "$1" != +(* OFFSET *) ]]
    then
	echo -e "\nERROR: Missing OFFSET in SEEK statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the variablename without surrounding spaces
    FILE=`echo ${1%% OFFSET *}`;

    # Get filedescriptor
    OFFSET=`echo ${1##* OFFSET }`

    WHENCE=

    # Check if WHENCE is available
    if [[ "$1" = +(* WHENCE *) ]]
    then
	WHENCE=${OFFSET##* WHENCE }
	OFFSET=${OFFSET%% WHENCE *}
    fi

    # Convert to C function
    case $WHENCE in
	+(START) )
	    echo "fseek($FILE, $OFFSET, SEEK_SET);" >> $g_CFILE;;
	+(CURRENT) )
	    echo "fseek($FILE, $OFFSET, SEEK_CUR);" >> $g_CFILE;;
	+(END) )
	    echo "fseek($FILE, $OFFSET, SEEK_END);" >> $g_CFILE;;
	*)
	    echo "fseek($FILE, $OFFSET, SEEK_SET);" >> $g_CFILE;;
    esac
}

#-----------------------------------------------------------

function Handle_Copy
{
    local FROM TO

    # Check if we have an argument at all
    if [[ "$1" = "COPY" ]]
    then
	echo -e "\nERROR: empty COPY at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in COPY statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the filename and copyname
    FROM=`echo ${1%% TO *}`
    TO=`echo ${1##* TO }`

    #translate to C function
    echo "FILE *__b2c__inFile; FILE *__b2c__outFile; int __b2c__Byte = 0;" >> $g_CFILE
    echo "__b2c__inFile = fopen($FROM, \"r\"); __b2c__outFile = fopen($TO, \"w\");" >> $g_CFILE
    echo "if (__b2c__inFile != NULL && __b2c__outFile != NULL){while(__b2c__Byte!=EOF){" >> $g_CFILE
    echo "__b2c__Byte=fgetc(__b2c__inFile); if(__b2c__Byte!=EOF){" >> $g_CFILE
    echo "fputc(__b2c__Byte,__b2c__outFile); }}}" >> $g_CFILE
    echo "fclose(__b2c__inFile); fclose(__b2c__outFile);" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Rename
{
    local FROM TO

    # Check if we have an argument at all
    if [[ "$1" = "RENAME" ]]
    then
	echo -e "\nERROR: empty RENAME at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi
    
    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in RENAME statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the filename and copyname
    FROM=`echo ${1%% TO *}`
    TO=`echo ${1##* TO }`

    # Translate to C function
    echo "if(rename($FROM, $TO) < 0) {if(!__b2c__trap){ERROR = 9; if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Color
{
    local TO FROM

    # Check if we have an argument at all
    if [[ "$1" = "COLOR" ]]
    then
	echo -e "\nERROR: empty COLOR at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if we need to reset
    if [[ "$1" = +(*RESET*) ]]
    then
	echo "printf(\"\033[0m\");" >> $g_CFILE
    elif [[ "$1" = +(*INTENSE*) ]]
    then
	echo "printf(\"\033[1m\");" >> $g_CFILE
    elif [[ "$1" = +(*INVERSE*) ]]
    then
	echo "printf(\"\033[7m\");" >> $g_CFILE
    elif [[ "$1" = +(*NORMAL*) ]]
    then
	echo "printf(\"\033[22m\");" >> $g_CFILE
    # Check if TO is available
    elif [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in COLOR statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    else
	# Get the target and colorname
	FROM=`echo ${1%% TO *}`
	TO=`echo ${1##* TO }`

	# Now select color
	if [[ $FROM = +(*FG*) ]]
	then
	    case $TO in
		"BLACK")
		    echo "printf(\"\033[30m\");" >> $g_CFILE;;
		"RED")
		    echo "printf(\"\033[31m\");" >> $g_CFILE;;
		"GREEN")
		    echo "printf(\"\033[32m\");" >> $g_CFILE;;
		"YELLOW")
		    echo "printf(\"\033[33m\");" >> $g_CFILE;;
		"BLUE")
		    echo "printf(\"\033[34m\");" >> $g_CFILE;;
		"MAGENTA")
		    echo "printf(\"\033[35m\");" >> $g_CFILE;;
		"CYAN")
		    echo "printf(\"\033[36m\");" >> $g_CFILE;;
		"WHITE")
		    echo "printf(\"\033[37m\");" >> $g_CFILE;;
	    esac
	elif [[ $FROM = +(*BG*) ]]
	then
	    case $TO in
		"BLACK")
		    echo "printf(\"\033[40m\");" >> $g_CFILE;;
		"RED")
		    echo "printf(\"\033[41m\");" >> $g_CFILE;;
		"GREEN")
		    echo "printf(\"\033[42m\");" >> $g_CFILE;;
		"YELLOW")
		    echo "printf(\"\033[43m\");" >> $g_CFILE;;
		"BLUE")
		    echo "printf(\"\033[44m\");" >> $g_CFILE;;
		"MAGENTA")
		    echo "printf(\"\033[45m\");" >> $g_CFILE;;
		"CYAN")
		    echo "printf(\"\033[46m\");" >> $g_CFILE;;
		"WHITE")
		    echo "printf(\"\033[47m\");" >> $g_CFILE;;
	    esac
	fi
    fi
}

#-----------------------------------------------------------

function Handle_Gotoxy
{
    local X Y

    # Check if we have an argument at all
    if [[ "$1" = "GOTOXY" ]]
    then
	echo -e "\nERROR: empty GOTOXY at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the target and colorname
    if [[ "$1" = +(*,*) ]]
    then
	X=`echo ${1%%,*}`
	Y=`echo ${1##*,}`
    else
	echo -e "\nERROR: missing coordinate in GOTOXY at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Translate to C
    echo "printf(\"\033[${Y};${X}H\");" >> $g_CFILE
}

#-----------------------------------------------------------

function Handle_Split
{
    typeset SOURCE REST BY TO SIZE TARGET CHECK

    # Check if we have an argument at all
    if [[ "$1" = "SPLIT" ]]
    then
	echo -e "\nERROR: empty SPLIT at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi
    
    # Check if BY is available
    if [[ "$1" != +(* BY *) ]]
    then
	echo -e "\nERROR: Missing BY in SPLIT statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if TO is available
    if [[ "$1" != +(* TO *) ]]
    then
	echo -e "\nERROR: Missing TO in SPLIT statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Check if SIZE is available
    if [[ "$1" != +(* SIZE *) ]]
    then
	echo -e "\nERROR: Missing SIZE in SPLIT statement at line $g_COUNTER in file $g_CURFILE!"
	exit 1
    fi

    # Get the source string
    SOURCE=`echo ${1%% BY *}`
    REST=`echo ${1##* BY }`

    BY=`echo ${REST%% TO *}`
    REST=`echo ${REST##* TO }`

    TO=`echo ${REST%% SIZE *}`
    SIZE=`echo ${REST##* SIZE }`

    # Determine target
    if [[ -n $g_FUNCNAME ]]
    then
	TARGET="$g_CFILE $g_HFILE"
    else
	TARGET=$g_HFILE
    fi

    touch $TARGET

    # Variable may not be array, these should be defined with DECLARE
    if [[ "$SIZE" != +(*\[*\]*) && "$SIZE" != +(*.*) ]]
    then
	# Declare variable if not done yet, assuming long
	CHECK=`grep -E "DIR|FILE|int|long|float|double|char|void|STRING|NUMBER|FLOATING" $TARGET | grep -E " $SIZE,| $SIZE;|,$SIZE,|,$SIZE;| ${SIZE%%\[*}\[|,${SIZE%%\[*}\[" | grep -v "\/\*record\*\/"`
	if [[ -z $CHECK ]]
	then
	    echo "long $SIZE;" >> $g_HFILE
	fi
    fi

    # Translate to C code
    if [[ -n $g_FUNCNAME ]]
    then
	if [[ $g_STRINGARGS != +(*\*\*$TO*) ]]
	then
	    g_STRINGARGS="$g_STRINGARGS char **$TO = NULL; int __b2c__split__$TO = 0;" 
	fi
    else
	if [[ -z `grep "$TO = NULL;" $g_HFILE` ]]
	then
	    echo "char **$TO = NULL; int __b2c__split__$TO = 0;" >> $g_HFILE
	fi
    fi

    # If the split array was used before in a loop, clear it
    echo "if($TO != NULL) {for(__b2c__ctr=0; __b2c__ctr<=__b2c__split__$TO; __b2c__ctr++) if($TO[__b2c__ctr]!=NULL)" >> $g_CFILE
    echo "{free($TO[__b2c__ctr]);} free($TO); $TO = NULL; __b2c__split__$TO = 0;}" >> $g_CFILE

    # Run the SPLIT code
    echo "if ($SOURCE != NULL && $BY != NULL && strlen($BY) > 0) {__b2c__split_tmp = strdup($SOURCE); __b2c__split = strtok(__b2c__split_tmp, $BY);" >> $g_CFILE
    echo "if (__b2c__split != NULL) {$TO = (char **)realloc($TO, (__b2c__split__$TO + 1) * sizeof(char*)); $TO[__b2c__split__$TO++] = strdup(__b2c__split);" >> $g_CFILE
    echo "while((__b2c__split = strtok(NULL, $BY)) != NULL) {$TO = (char **)realloc($TO, (__b2c__split__$TO + 1) * sizeof(char*));" >> $g_CFILE
    echo "$TO[__b2c__split__$TO++] = strdup(__b2c__split);} } free(__b2c__split_tmp);} $SIZE = __b2c__split__$TO--;" >> $g_CFILE

    # Add declared array to array list if we are in a function
    if [[ -n $g_FUNCNAME && $g_STRINGARRAYS != +(*$TO\[*) ]]
    then
	g_STRINGARRAYS="$g_STRINGARRAYS if($TO != NULL) {for(__b2c__ctr=0; __b2c__ctr<=__b2c__split__$TO; __b2c__ctr++) if($TO[__b2c__ctr]!=NULL){free($TO[__b2c__ctr]);} free($TO);}"
    fi
}

#-----------------------------------------------------------
#
# Simple parser to tokenize line.
#
# Each line should begin with a statement.
# The rest of the line may contain functions, these are
#   converted using C macros.
#
#-----------------------------------------------------------

function Parse_Line
{
    local FOUND SYM INC COPY_COUNTER COPY_CURFILE STATEMENT
    local LEN SEQ TOTAL

    # Get statement without spaces
    STATEMENT=`echo "${1%% *}"`

    if [[ $g_USE_C -eq 1 ]]
    then
	if [[ "$STATEMENT" = "END" || "$STATEMENT" = "ENDUSEC" ]]
	then
	    let g_USE_C=0
	else
	    echo "${1}" >>  $g_CFILE
	fi
    else
	case "$STATEMENT" in
	    "USEC")
		let g_USE_C=1;;
	    "PRINT")
		Handle_Print "${1#* }";;
	    "INPUT")
		Handle_Input "${1#* }";;
	    "FOR")
		Handle_For "${1#* }";;
	    "NEXT")
		echo "}" >> $g_CFILE;;
	    "IF")
		Handle_If "${1#* }";;
	    "ELIF")
		Handle_Elif "${1#* }";;
	    "ELSE")
		echo "} else {" >> $g_CFILE;;
	    "ENDIF")
		echo "}" >> $g_CFILE;;
	    "WHILE")
		Handle_While "${1#* }";;
	    "WEND")
		echo "}" >> $g_CFILE;;
	    "BREAK")
		echo "break;" >> $g_CFILE;;
	    "REPEAT")
		echo "do{" >> $g_CFILE;;
	    "UNTIL")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty UNTIL at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Convert to legal C code
		    echo "} while(!(${1#* }));" >> $g_CFILE
		fi;;
	    "LET")
		Handle_Let "${1#* }";;
	    +(REM*) )
		;;
	    +(\'*) )
		;;
	    "SYSTEM")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty SYSTEM at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    echo "SYSTEM (${1#* });" >> $g_CFILE
		fi;;
	    "SLEEP")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty SLEEP at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    echo "usleep(${1#* }*1000);" >> $g_CFILE
		fi;;
	    "OPEN")
		Handle_Open "${1#* }";;
	    "CLOSE")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty CLOSE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    SYM="${1#* }"
		    if [[ "$SYM" = +(*FILE*) ]]
		    then
			echo "fclose(${1##* });" >> $g_CFILE
		    elif [[ "$SYM" = +(*DIRECTORY*) ]]
		    then
			echo "closedir(${1##* });" >> $g_CFILE
		    elif [[ "$SYM" = +(*MEMORY*) ]]
		    then
			echo "${1##* } = NULL;" >> $g_CFILE
		    elif [[ "$SYM" = +(*NETWORK*) || "$SYM" = +(*SERVER*) ]]
		    then
			echo "close(${1##* });" >> $g_CFILE
		    else
			echo -e "\nERROR: erronuous CLOSE argument at line $g_COUNTER in file $g_CURFILE!"
			exit 1
		    fi
		fi;;
	    "REWIND")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty REWIND at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    echo "rewind(${1#* });" >> $g_CFILE
		fi;;
	    "MEMREWIND")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty MEMREWIND at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    echo "${1#* } = (char*)__b2c_mem_${1#* };" >> $g_CFILE
		fi;;
	    "SEEK")
		Handle_Seek "${1#* }";;
	    "READLN")
		Handle_Readln "${1#* }";;
	    "WRITELN")
		Handle_Writeln "${1#* }";;
	    "GETBYTE")
		Handle_Getbyte "${1#* }";;
	    "PUTBYTE")
		Handle_Putbyte "${1#* }";;
	    "GETFILE")
		Handle_Getfile "${1#* }";;
	    "GETLINE")
		Handle_Getline "${1#* }";;
	    "PUTLINE")
		Handle_Putline "${1#* }";;
	    "END")
		if [[ "${1#* }" = +(*IF*) ]]
		then
		    echo "}" >> $g_CFILE
		elif [[ "${1#* }" = +(*RECORD*) ]]
		then
		    if [[ -n $g_FUNCNAME ]]
		    then
			echo "};" >> $g_CFILE
			echo "struct $g_RECORDNAME $g_RECORDVAR = { } ;" >> $g_CFILE
		    else
			echo "};" >> $g_HFILE
			echo "struct $g_RECORDNAME $g_RECORDVAR = { } ;" >> $g_HFILE
		    fi
		    g_RECORDNAME=
		    g_RECORDVAR=
		elif [[ "${1#* }" = +(*FUNCTION*) ]]
		then
		    Handle_Endfunction
		elif [[ "${1#* }" = +(*SUB*) ]]
		then
		    Handle_Endsub
		elif [[ "${1#* }" = +(*WITH*) ]]
		then
		    g_WITHVAR=
		elif [[ "${1#* }" != "END" ]]
		then
		    echo "exit(${1#* });" >> $g_CFILE
		else
		    echo "exit(EXIT_SUCCESS);" >> $g_CFILE
		fi;;
	    "SUB")
		Handle_SubFunc "${1#* }";;
	    "ENDSUB")
		Handle_Endsub;;
	    "CALL")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty CALL at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    SYM=`echo "${1#* }"`
		    if [[ "$SYM" = +(* TO *) ]]
		    then
			echo "${SYM##*TO } = " >> $g_CFILE
			if [[ "${SYM%% TO*}" = +(*\(*\)*) ]]
			then
			    echo "${SYM%% TO*};" >> $g_CFILE
			else
			    echo "${SYM%% TO*}();" >> $g_CFILE
			fi
		    elif [[ "$SYM" = +(*\(*\)*) ]]
		    then
			echo "$SYM;" >> $g_CFILE
		    else
			echo "$SYM();" >> $g_CFILE
		    fi
		fi;;
	    "FUNCTION")
		Handle_SubFunc "${1#* }";;
	    "ENDFUNCTION")
		Handle_Endfunction;;
	    "RETURN")
		Handle_Return "${1#* }";;
	    "IMPORT")
		Handle_Import "${1#* }";;
	    "DECLARE")
		Handle_Declare "${1#* }";;
	    "GLOBAL")
		Handle_Declare "${1#* }";;
	    "LOCAL")
		Handle_Local "${1#* }";;
	    "DATA")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty DATA at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    if [[ "${1#* }" != `echo "${1#* }" | tr -d "\042"` ]]
		    then
			echo "${1#* }, " >> $STRINGARRAYFILE
		    else
			# Check if float
			if [[ "${1#* }" = +(*.*) ]]
			then
			    echo "${1#* }, " >> $FLOATARRAYFILE
			else
			    echo "${1#* }, " >> $INTARRAYFILE
			fi
		    fi
		fi;;
	    "RESTORE")
		echo "if(__b2c__intarray_ptr > 0) __b2c__intarray_ptr = 0;" >> $g_CFILE
		echo "if(__b2c__floatarray_ptr > 0) __b2c__floatarray_ptr = 0;" >> $g_CFILE
		echo "if(__b2c__stringarray_ptr > 0) __b2c__stringarray_ptr = 0;" >> $g_CFILE;;
	    "READ")
		Handle_Read "${1#* }";;
	    "PUSH")
		Handle_Push "${1#* }";;
	    "PULL")
		Handle_Pull "${1#* }";;
	    "SEED")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty SEED at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    echo "srandom((unsigned int)${1#* });" >> $g_CFILE
		fi;;
	    "DEF")
		Handle_Deffn "${1#* }";;
	    "CONST")
		Handle_Const "${1#* }";;
	    "INCLUDE")
		# Get rid of doublequotes if they are there
		INC=`echo ${1#* } | tr -d "\042"`

		# Check argument
		if [[ ! -f $INC || "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: missing file for INCLUDE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Start walking through program
		    COPY_COUNTER=$g_COUNTER
		    g_COUNTER=1
		    # Assign new file
		    COPY_CURFILE=$g_CURFILE
		    g_CURFILE=$INC
		    while read -r LINE
		    do
			echo -e -n "\rStarting conversion... $g_COUNTER"
			# Line is not empty?
			if [[ -n "$LINE" ]]
			then
			    if [[ "$LINE" = +(* \\) ]]
			    then
				let LEN="${#LINE}"-1
				SEQ="${LINE:0:$LEN}"
				TOTAL=$TOTAL$SEQ
			    else
				Parse_Line "$TOTAL$LINE"
				TOTAL=
			    fi
			fi
			((g_COUNTER+=1))
		    done < $INC
		    # Restore original counter
		    g_COUNTER=$COPY_COUNTER
		    # Restore original file
		    g_CURFILE=$COPY_CURFILE
		fi;;
	    "POKE")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty POKE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Resolve this with C macro because of casting to (char)
		    echo "POKE(${1#* });" >> $g_CFILE
		fi;;
	    "RESIZE")
	    	# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty RESIZE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Resolve this with C macro because of casting to (char)
		    echo "RESIZE(${1#* });" >> $g_CFILE
		fi;;
	    "COPY")
		Handle_Copy "${1#* }";;
	    "DELETE")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty DELETE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C function
		    if [[ "${1#* }" = +(*FILE*) ]]
		    then
			echo "if (unlink(${1#*FILE })==-1){if(!__b2c__trap){ERROR = 7;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
		    elif [[ "${1#* }" = +(*DIRECTORY*) ]]
		    then
			 echo "if (rmdir(${1#*DIRECTORY }) == -1){if(!__b2c__trap){ERROR = 20;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
		    else
			echo "\nERROR: erronuous argument for DELETE at line $g_COUNTER in file $g_CURFILE!"
			exit 1
		    fi
		fi;;
	    "RENAME")
		Handle_Rename "${1#* }";;
	    "MAKEDIR")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty MAKEDIR at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C function
		    echo "if (mkdir(${1#* }, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1){if(!__b2c__trap){ERROR = 21;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
		fi;;
	    "CHANGEDIR")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty CHANGEDIR at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C function
		    echo "if (chdir(${1#* }) == -1){if(!__b2c__trap) {ERROR = 22;if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} }" >> $g_CFILE
		fi;;
	    "FREE")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty FREE at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C function
		    echo "if(!__b2c__trap){__b2c__memory__check((char *)${1#* });" >> $g_CFILE
		    echo "if(ERROR) {if(!__b2c__catch_set){fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} if(!setjmp(__b2c__jump)) goto $g_CATCHGOTO;} } free((void*)${1#* });" >> $g_CFILE
		fi;;
	    "GOTO")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty GOTO at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C label
		    echo "goto ${1#* };" >> $g_CFILE
		fi;;
	    "LABEL")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty LABEL at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    # Translate to C label
		    echo "${1#* }:" >> $g_CFILE
		    echo ";" >> $g_CFILE
		fi;;
	    "TRAP")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty TRAP at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		else
		    if [[ "${1#* }" = "SYSTEM" ]]
		    then
			echo "/* Error catching is enabled */" >> $g_CFILE
			echo "__b2c__trap = 1;" >> $g_CFILE
			echo "signal(SIGILL, __b2c__catch_signal);" >> $g_CFILE
			echo "signal(SIGABRT, __b2c__catch_signal);" >> $g_CFILE
			echo "signal(SIGFPE, __b2c__catch_signal);" >> $g_CFILE
			echo "signal(SIGSEGV, __b2c__catch_signal);" >> $g_CFILE
		    elif [[ "${1#* }" = "LOCAL" ]]
		    then
			echo "/* Error catching is disabled */" >> $g_CFILE
			echo "__b2c__trap = 0;" >> $g_CFILE
			echo "signal(SIGILL, SIG_DFL);" >> $g_CFILE
			echo "signal(SIGABRT, SIG_DFL);" >> $g_CFILE
			echo "signal(SIGFPE, SIG_DFL);" >> $g_CFILE
			echo "signal(SIGSEGV, SIG_DFL);" >> $g_CFILE
		    else
			echo -e "\nERROR: invalid argument for TRAP at line $g_COUNTER in file $g_CURFILE!"
			exit 1
		    fi
		fi;;
	    "CATCH")
		# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty CATCH at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		elif [[ "${1#* }" = +(*GOTO*) ]]
		then
		    echo "__b2c__catch_set = 1;" >> $g_CFILE
		    g_CATCHGOTO="${1##* GOTO }"
		elif [[ "${1#* }" = +(*RESET*) ]]
		then
		    echo "__b2c__catch_set = 0;" >> $g_CFILE
		else
		    echo -e "\nERROR: CATCH without GOTO or RESET at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		fi;;
	    "RESUME")
		echo "longjmp(__b2c__jump, 1);" >> $g_CFILE;;
	    "CLEAR")
		echo "printf(\"\033[2J\"); printf(\"\033[0;0f\");" >> $g_CFILE;;
	    "COLOR")
		Handle_Color "${1#* }";;
	    "GOTOXY")
		Handle_Gotoxy "${1#* }";;
	    "RECEIVE")
		Handle_Receive "${1#* }";;
	    "SEND")
		Handle_Send "${1#* }";;
	    "RECORD")
	    	# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    echo -e "\nERROR: empty RECORD at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		elif [[ -n $g_RECORDNAME ]]
		then
		    echo -e "\nERROR: cannot define a record within a record!"
		    exit 1
		else
		    # Translate to C typedef struct
		    if [[ -n $g_FUNCNAME ]]
		    then
			g_RECORDNAME="RECORD_${g_FUNCNAME}_${g_COUNTER}"
			echo "struct $g_RECORDNAME{" >> $g_CFILE
		    else
			g_RECORDNAME="RECORD_${g_COUNTER}"
			echo "struct $g_RECORDNAME{" >> $g_HFILE
		    fi
		    g_RECORDVAR=${1#* }
		fi;;
	    "WITH")
	    	# Check argument
		if [[ "${1%% *}" = "${1#* }" ]]
		then
		    print "\nERROR: empty WITH at line $g_COUNTER in file $g_CURFILE!"
		    exit 1
		elif [[ -n $g_RECORDNAME ]]
		then
		    print "\nERROR: cannot define a WITH within a WITH!"
		    exit 1
		else
		    g_WITHVAR=${1#* }
		fi;;
	    "SPLIT")
		Handle_Split "${1#* }";;
	    *)
		# Check on imported symbols first
		FOUND=0
		for SYM in $g_IMPORTED
		do
		    if [[ "$SYM" = ${1%%\(*} || "$SYM" = ${1%% *} ]]
		    then
			FOUND=1
			if [[ "$1" != +(*\(*\)*) ]]
			then
			    echo "$SYM();" >> $g_CFILE
			else
			    echo "$SYM(${1#*\(};" >> $g_CFILE
			fi
		    fi
		done
		# Not an imported symbol? Check if assignment
		if [[ $FOUND -eq 0 ]]
		then
		    Handle_Let "$1"
		fi;;
	esac
    fi
}

#-----------------------------------------------------------
#
# Main program
#
#-----------------------------------------------------------

# Default BACON settings
let g_MAX_DIGITS=32

# Maximum of internal buffers needed for string processing
let g_MAX_BUFFERS=32

# This is the size for static buffers like fgets, read etc.
let g_BUFFER_SIZE=512

# Maximum RETURN buffers
let g_MAX_RBUFFERS=8

let g_NO_COMPILE=0
let g_STACK_DEPTH=8
let g_TMP_PRESERVE=0
let g_USE_C=0
let g_MAX_BACKLOG=4

g_TEMPDIR=.
g_CCNAME=gcc
g_CCFLAGS=
g_INCFILES=
g_INCLUDE_FILES=

# Some global declarations
g_CURFILE=
g_FUNCNAME=
g_FUNCTYPE=
g_TMP_FILES=
g_LOCALSTRINGS=
g_STRINGARRAYS=
g_STRINGARGS=

# Always create a final label
g_CATCHGOTO="__B2C__PROGRAM__EXIT"

# Records
g_RECORDNAME=
g_RECORDVAR=
g_WITHVAR=

# Get arguments
while getopts ":m:c:s:d:i:l:awognpv" OPT
do
    case $OPT in
	m) g_MAX_BUFFERS=$OPTARG;;
	c) g_CCNAME=$OPTARG;;
	s) g_STACK_DEPTH=$OPTARG;;
	d) g_TEMPDIR=$OPTARG;;
	i) g_INCFILES="$g_INCFILES $OPTARG";;
	l) g_LDFLAGS="$g_LDFLAGS -l$OPTARG";;
	n) g_NO_COMPILE=1;;
	g) g_CCFLAGS="$g_CCFLAGS -g";; 
	w) g_CCFLAGS="$g_CCFLAGS -Wall";; 
	o) g_CCFLAGS="$g_CCFLAGS -mtune=native";; 
	p) let g_TMP_PRESERVE=1;;
	a) g_CCFLAGS="$g_CCFLAGS -m32";;
	v) echo 
	   echo "BACON version $g_VERSION - (c) Peter van Eerten - GPL."
	   echo 
	   exit 0;; 
	\?|h) echo 
	    echo "USAGE: bacon [options] program[.bac]"
	    echo 
	    echo "OPTIONS:"
	    echo 
	    echo -e " -c <compiler>\tCompiler to use (default: gcc)"
	    echo -e " -i <include>\tAdd include file to C code"
	    echo -e " -l <library>\tPass library to linker"
	    echo -e " -m <value>\tMax temporary buffers (default: 32)"
	    echo -e " -s <value>\tDepth of stack (default: 8)"
	    echo -e " -d <tmpdir>\tTemporary directory (default: .)"
	    echo -e " -n \t\tDo not compile automatically"
	    echo -e " -p \t\tPreserve temporary files"
	    echo -e " -o \t\tOptimize binary for current platform"
	    echo -e " -a \t\tForce 32bit compilation"
	    echo -e " -g \t\tCompile with debug info"
	    echo -e " -w \t\tShow all warnings during compilation"
	    echo -e " -v \t\tShow version"
	    echo -e " -h \t\tShow help"
	    echo 
	    exit 0;;
    esac
done

shift $(($OPTIND-1))

# Check if a filename was entered, if so get it
if [[ $# -eq 0 ]]
then
    echo -e "ERROR: no filename? Run with '-h' to see usage."
    exit 1
elif [[ "$@" != +(*.bac) ]]
then
    SOURCEFILE="$@.bac"
else
    SOURCEFILE="$@"
fi

# Check if file exists
if [[ ! -f $SOURCEFILE ]]
then
    echo -e "ERROR: file not found!"
    exit 1
fi

echo -n "Starting conversion... "

# Now create the global filenames where to write to
g_CFILE=$g_TEMPDIR/${SOURCEFILE}.c
g_HFILE=$g_TEMPDIR/${SOURCEFILE}.h
STRINGARRAYFILE=$g_TEMPDIR/${SOURCEFILE}.string.h
INTARRAYFILE=$g_TEMPDIR/${SOURCEFILE}.int.h
FLOATARRAYFILE=$g_TEMPDIR/${SOURCEFILE}.float.h

# Add to total file list
g_TMP_FILES="$g_CFILE $g_HFILE $STRINGARRAYFILE $INTARRAYFILE $FLOATARRAYFILE"

# Create basic C file
echo "/* Created with BaCon $g_VERSION - Peter van Eerten 2009 */" > $g_CFILE
echo "#include \"$g_HFILE\"" >> $g_CFILE
echo "int main(int argc, const char **argv)" >> $g_CFILE
echo "{" >> $g_CFILE
echo "/* Default is: systems traps signals */" >> $g_CFILE
echo "signal(SIGILL, __b2c__catch_signal);" >> $g_CFILE
echo "signal(SIGABRT, __b2c__catch_signal);" >> $g_CFILE
echo "signal(SIGFPE, __b2c__catch_signal);" >> $g_CFILE
echo "signal(SIGSEGV, __b2c__catch_signal);" >> $g_CFILE
echo "/* Make sure internal string buffers are empty */" >> $g_CFILE
echo "for(__b2c__sbuffer_ptr=0; __b2c__sbuffer_ptr < $g_MAX_BUFFERS; __b2c__sbuffer_ptr++) __b2c__sbuffer[__b2c__sbuffer_ptr] = NULL;" >> $g_CFILE
echo "__b2c__sbuffer_ptr=0;" >> $g_CFILE
echo "for(__b2c__stackptr=0; __b2c__stackptr < $g_STACK_DEPTH; __b2c__stackptr++) __b2c__stringstack[__b2c__stackptr] = NULL;" >> $g_CFILE
echo "__b2c__stackptr=0;" >> $g_CFILE
echo "for(__b2c__rbuffer_ptr=0; __b2c__rbuffer_ptr < $g_MAX_BUFFERS; __b2c__rbuffer_ptr++) __b2c__rbuffer[__b2c__rbuffer_ptr] = NULL;" >> $g_CFILE
echo "__b2c__rbuffer_ptr=0;" >> $g_CFILE

# Put arguments into reserved variable ARGUMENT
echo "/* Setup the reserved variable 'ARGUMENT' */" >> $g_CFILE
echo "for(__b2c__counter=1; __b2c__counter < argc; __b2c__counter++)" >> $g_CFILE
echo "{__b2c__arglen += strlen(argv[__b2c__counter]) + 1;} __b2c__arglen++;" >> $g_CFILE
echo "ARGUMENT\$ = (char*)malloc(__b2c__arglen*sizeof(char)); strcpy(ARGUMENT\$, \"\");" >> $g_CFILE
echo "for(__b2c__counter=1; __b2c__counter < argc; __b2c__counter++)" >> $g_CFILE
echo "{strcat(ARGUMENT\$, argv[__b2c__counter]); if(__b2c__counter != argc - 1) strcat(ARGUMENT\$, \" \");}" >> $g_CFILE
echo "/* Rest of the program */" >> $g_CFILE

# Create basic H file, functions are converted using macros
echo "/* Created with BaCon $g_VERSION - Peter van Eerten 2009 */" > $g_HFILE
echo >> $g_HFILE
echo "#include <stdio.h>" >> $g_HFILE
echo "#include <stdlib.h>" >> $g_HFILE
echo "#include <stdarg.h>" >> $g_HFILE
echo "#include <sys/time.h>" >> $g_HFILE
echo "#include <sys/stat.h>" >> $g_HFILE
echo "#include <sys/types.h>" >> $g_HFILE
echo "#include <sys/socket.h>" >> $g_HFILE
echo "#include <sys/utsname.h>" >> $g_HFILE
echo "#include <dirent.h>" >> $g_HFILE
echo "#include <setjmp.h>" >> $g_HFILE
echo "#include <netdb.h>" >> $g_HFILE
if [[ `uname` = +(*BSD*) ]]
then
    echo "#include <netinet/in.h>" >> $g_HFILE
fi
echo "#include <signal.h>" >> $g_HFILE
echo "static jmp_buf __b2c__jump;" >> $g_HFILE
echo "static int __b2c__trap = 1;" >> $g_HFILE
echo "static int __b2c__catch_set = 0;" >> $g_HFILE
echo "static int ERROR = 0;" >> $g_HFILE
echo "const double VERSION = $g_VERSION;" >> $g_HFILE
echo >> $g_HFILE
# Add user include files
echo "/* User include files */" >> $g_HFILE
for i in $g_INCFILES
do
    if [[ "$i" = +(*<*) ]]
    then
	echo "#include $i" >> $g_HFILE
    else
	echo "#include \"$i\"" >> $g_HFILE
    fi
done
echo "/* READ/DATA include files */" >> $g_HFILE
echo "int __b2c__stringarray_ptr = 0;" >> $g_HFILE
echo "#include \"$STRINGARRAYFILE\"" >> $g_HFILE
echo "int __b2c__intarray_ptr = 0;" >> $g_HFILE
echo "#include \"$INTARRAYFILE\"" >> $g_HFILE
echo "int __b2c__floatarray_ptr = 0;" >> $g_HFILE
echo "#include \"$FLOATARRAYFILE\"" >> $g_HFILE
echo "int __b2c__ctr;" >> $g_HFILE
echo >> $g_HFILE
echo "/* Math functions */" >> $g_HFILE
echo "extern double round(double __b2c__x);" >> $g_HFILE
echo "extern long int lrint(double __b2c__x);" >> $g_HFILE
echo "#include <math.h>" >> $g_HFILE
echo "#define SQR(__b2c__x) sqrt(__b2c__x)" >> $g_HFILE
echo "#define POW(__b2c__x, __b2c__y) pow(__b2c__x, __b2c__y)" >> $g_HFILE
echo "#define SIN(__b2c__x) sin(__b2c__x)" >> $g_HFILE
echo "#define COS(__b2c__x) cos(__b2c__x)" >> $g_HFILE
echo "#define TAN(__b2c__x) tan(__b2c__x)" >> $g_HFILE
echo "#define ROUND(__b2c__x) lrint(__b2c__x)" >> $g_HFILE
echo "#define INT(__b2c__x) lrint(__b2c__x)" >> $g_HFILE
echo "#define MOD(__b2c__x, __b2c__y) (__b2c__x % __b2c__y)" >> $g_HFILE
echo "#define EVEN(__b2c__x) ((__b2c__x % 2 == 0) ? 1 : 0)" >> $g_HFILE
echo "#define ODD(__b2c__x) ((__b2c__x % 2 != 0) ? 1 : 0)" >> $g_HFILE
echo "#define FLOOR(__b2c__x) (long)floor(__b2c__x)" >> $g_HFILE
echo "#define ABS(__b2c__x) abs(__b2c__x)" >> $g_HFILE
echo "#define RND random()" >> $g_HFILE
echo "#define MAXRANDOM RAND_MAX" >> $g_HFILE
echo >> $g_HFILE
echo "/* Other functions */" >> $g_HFILE
echo "#define VAL(__b2c__x) ((__b2c__x != NULL) ? atof(__b2c__x) : 0)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Unix functions */" >> $g_HFILE
echo "#include <unistd.h>" >> $g_HFILE
echo "#define SYSTEM(__b2c__x) ((__b2c__x != NULL) ? system(__b2c__x) : 0)" >> $g_HFILE
echo >> $g_HFILE
echo "/* String functions */" >> $g_HFILE
echo "#include <string.h>" >> $g_HFILE
echo "char __b2c__input__buffer[$g_BUFFER_SIZE];" >> $g_HFILE
echo "char* __b2c__sbuffer[$g_MAX_BUFFERS];" >> $g_HFILE
echo "int __b2c__sbuffer_ptr;" >> $g_HFILE
echo "char* __b2c__rbuffer[$g_MAX_RBUFFERS];" >> $g_HFILE
echo "int __b2c__rbuffer_ptr;" >> $g_HFILE
echo "/* Temporary pointer to perform assignments */" >> $g_HFILE
echo "char *__b2c__assign = NULL;" >> $g_HFILE
echo "char *__b2c__split = NULL; char *__b2c__split_tmp = NULL;" >> $g_HFILE
echo "char *ERR\$(int);" >> $g_HFILE
echo "/* Actual functions */" >> $g_HFILE
echo "char* __b2c__reverse(char *__b2c__s){int __b2c__i; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__s)+2)*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, strlen(__b2c__s)+1);" >> $g_HFILE
echo "for(__b2c__i=0;__b2c__i<strlen(__b2c__s);__b2c__i++){__b2c__sbuffer[__b2c__sbuffer_ptr][__b2c__i]=__b2c__s[strlen(__b2c__s)-1-__b2c__i];}" >> $g_HFILE
echo "return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define REVERSE\$(x) ((x != NULL) ? __b2c__reverse(x) : \"null\")" >> $g_HFILE
echo "char* __b2c__str(double d){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, $g_MAX_DIGITS);" >> $g_HFILE
echo "if(floor(d) == d) snprintf(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%ld\", (long)d); else snprintf(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%g\", d);" >> $g_HFILE
echo "return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define STR\$(x) __b2c__str(x)" >> $g_HFILE
echo "char* __b2c__concat(char *__b2c__first, ...){char *__b2c__tmp; va_list __b2c__ap; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "if(__b2c__first != NULL) { __b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__first)+1)*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, (strlen(__b2c__first)+1)*sizeof(char)); strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__first);}" >> $g_HFILE
echo "else {__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], 2*sizeof(char)); strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], \"\");} va_start(__b2c__ap, __b2c__first);" >> $g_HFILE
echo "while((__b2c__tmp = va_arg(__b2c__ap, char *)) != NULL) {__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__tmp)+strlen(__b2c__sbuffer[__b2c__sbuffer_ptr])+1)*sizeof(char));" >> $g_HFILE
echo "strcat(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__tmp);} va_end(__b2c__ap); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define CONCAT\$(...) __b2c__concat(__VA_ARGS__, (char*)NULL)" >> $g_HFILE
echo "char* __b2c__left(char *__b2c__src, long __b2c__n){if(__b2c__src == NULL || __b2c__n > strlen(__b2c__src) || __b2c__n < 0) return(__b2c__src);" >> $g_HFILE
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__src)+1)*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, (strlen(__b2c__src)+1)*sizeof(char)); strncpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__src, (__b2c__n));" >> $g_HFILE
echo "return (__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define LEFT\$(__b2c__x, __b2c__y) __b2c__left(__b2c__x, __b2c__y)" >> $g_HFILE
echo "char* __b2c__right(char *__b2c__src, long __b2c__n){if(__b2c__src == NULL || __b2c__n > strlen(__b2c__src) || __b2c__n < 0) return(__b2c__src);" >> $g_HFILE
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__src)+1)*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, (strlen(__b2c__src)+1)*sizeof(char)); strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__src+ strlen(__b2c__src)- __b2c__n);" >> $g_HFILE
echo "return (__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define RIGHT\$(__b2c__x, __b2c__y) __b2c__right(__b2c__x, __b2c__y)" >> $g_HFILE
echo "char* __b2c__mid(char *__b2c__src, ...){ va_list __b2c__ap; int __b2c__start, __b2c__end; if(__b2c__src == NULL) return(\"\");" >> $g_HFILE
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__src)+1)*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, (strlen(__b2c__src)+1)*sizeof(char)); va_start(__b2c__ap, __b2c__src);" >> $g_HFILE
echo "__b2c__start = va_arg(__b2c__ap, long); if(__b2c__start < 1) return(__b2c__src); __b2c__end = va_arg(__b2c__ap, long); if((__b2c__end) == -1) __b2c__end = strlen(__b2c__src);" >> $g_HFILE
echo "strncpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__src+(long)(__b2c__start)-1, (__b2c__end)); va_end(__b2c__ap); return (__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define MID\$(__b2c__string, ...) __b2c__mid(__b2c__string, __VA_ARGS__, -1)" >> $g_HFILE
echo "long __b2c__instr(char *__b2c__first, ...){char *__b2c__tmp; char *__b2c__res; int __b2c__pos; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "va_list __b2c__ap; if(__b2c__first == NULL) return (0); __b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__first)+1)*sizeof(char));" >> $g_HFILE
echo "strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__first); va_start(__b2c__ap, __b2c__first);" >> $g_HFILE
echo "__b2c__tmp = va_arg(__b2c__ap, char *); if(__b2c__tmp == NULL || strlen(__b2c__tmp) == 0) return (0);__b2c__pos = va_arg(__b2c__ap, int);" >> $g_HFILE
echo "if(__b2c__pos <= 0) __b2c__pos = 1; va_end(__b2c__ap); __b2c__res = strstr(__b2c__sbuffer[__b2c__sbuffer_ptr] + __b2c__pos - 1, __b2c__tmp);" >> $g_HFILE
echo "if(__b2c__res == NULL) return (0); return (__b2c__res - __b2c__sbuffer[__b2c__sbuffer_ptr] + 1);}" >> $g_HFILE
echo "#define INSTR(...) __b2c__instr(__VA_ARGS__, -1)" >> $g_HFILE
echo "long __b2c__instrrev(char *__b2c__first, ...){char *__b2c__tmp; char *__b2c__res = NULL; int __b2c__pos; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "va_list __b2c__ap; if(__b2c__first == NULL) return (0); __b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__first)+1)*sizeof(char));" >> $g_HFILE
echo "strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__first); va_start(__b2c__ap, __b2c__first);" >> $g_HFILE
echo "__b2c__tmp = va_arg(__b2c__ap, char *); if(__b2c__tmp == NULL || strlen(__b2c__tmp) == 0) return (0);__b2c__pos = va_arg(__b2c__ap, int);" >> $g_HFILE
echo "if(__b2c__pos <= 0) __b2c__pos = 1; va_end(__b2c__ap); while(__b2c__res == NULL && __b2c__pos < strlen(__b2c__sbuffer[__b2c__sbuffer_ptr])) {" >> $g_HFILE
echo "__b2c__res = strstr(__b2c__sbuffer[__b2c__sbuffer_ptr] + strlen(__b2c__sbuffer[__b2c__sbuffer_ptr]) - __b2c__pos + 1, __b2c__tmp); __b2c__pos+=1;}" >> $g_HFILE
echo "if(__b2c__res == NULL) return (0); return (__b2c__res - __b2c__sbuffer[__b2c__sbuffer_ptr] + 1);}" >> $g_HFILE
echo "#define INSTRREV(...) __b2c__instrrev(__VA_ARGS__, -1)" >> $g_HFILE
echo "char* __b2c__spc(int __b2c__x){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 32, __b2c__x); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define SPC\$(x) ((x > -1) ? __b2c__spc(x) : \"\")" >> $g_HFILE
echo "char* __b2c__tab(int __b2c__x){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 9, __b2c__x); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define TAB\$(x) ((x > -1) ? __b2c__tab(x) : \"\")" >> $g_HFILE
echo "char* __b2c__fill(unsigned long __b2c__x, unsigned char __b2c__y){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, __b2c__x*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__y, __b2c__x); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define FILL\$(x, y) ((y > -1 && y < 256) ? __b2c__fill(x, y) : \"\")" >> $g_HFILE
echo "char* __b2c__chop(char *__b2c__src){char *__b2c__tmp; int __b2c__i; if(strlen(__b2c__src) == 0) return(__b2c__src);" >> $g_HFILE
echo "while ((*__b2c__src == 10 || *__b2c__src == 13 || *__b2c__src == 9 || *__b2c__src == 32) && *__b2c__src != '\0') __b2c__src++;" >> $g_HFILE
echo "if (*__b2c__src == '\0') return(\"\"); __b2c__tmp = __b2c__src + strlen(__b2c__src) - 1;" >> $g_HFILE
echo "while ((*__b2c__tmp == 10 || *__b2c__tmp == 13 || *__b2c__tmp == 9 || *__b2c__tmp == 32) && *__b2c__src != '\0') __b2c__tmp--;" >> $g_HFILE
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__src)+1)*sizeof(char));" >> $g_HFILE
echo "for(__b2c__i = 0; __b2c__i <= __b2c__tmp - __b2c__src; __b2c__i++) __b2c__sbuffer[__b2c__sbuffer_ptr][__b2c__i]=__b2c__src[__b2c__i];" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr][__b2c__i] = '\0'; return (__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define CHOP\$(x) ((x != NULL) ? __b2c__chop(x) : 0)" >> $g_HFILE
echo "char* __b2c__replace(char* __b2c__x, char* __b2c__y, char* __b2c__z){char *__b2c__tmp, *__b2c__buf = NULL;" >> $g_HFILE
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__x)+1)*sizeof(char));" >> $g_HFILE
echo "strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__x); while(__b2c__tmp = strstr(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__y)) {" >> $g_HFILE
echo "__b2c__buf = malloc((strlen(__b2c__sbuffer[__b2c__sbuffer_ptr]) + strlen(__b2c__z) + 1)*sizeof(char)); *__b2c__tmp = '\0';" >> $g_HFILE
echo "strcpy(__b2c__buf, __b2c__sbuffer[__b2c__sbuffer_ptr]); strcat(__b2c__buf, __b2c__z); __b2c__tmp += strlen(__b2c__y); strcat(__b2c__buf, __b2c__tmp);" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], (strlen(__b2c__buf)+1)*sizeof(char));" >> $g_HFILE
echo "strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__buf); free(__b2c__buf);} return (__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define REPLACE\$(x, y, z)((x!=NULL && y!=NULL && z!= NULL) ? __b2c__replace(x, y, z) : 0)" >> $g_HFILE
echo "#define LEN(x) ((x != NULL) ? strlen(x) : 0)" >> $g_HFILE
echo "#define EQUAL(__b2c__x, __b2c__y) ((__b2c__x != NULL && __b2c__y != NULL) ? !strcmp(__b2c__x, __b2c__y) : 0)" >> $g_HFILE
echo "char * __b2c__getenv(char *__b2c__env){static char * __b2c__tmp; __b2c__tmp = getenv(__b2c__env); if (__b2c__tmp == NULL)" >> $g_HFILE
echo "{return \"\";} return __b2c__tmp;}" >> $g_HFILE
echo "#define GETENVIRON\$(x) ((x != NULL) ? __b2c__getenv(x) : \"null\")" >> $g_HFILE
echo >> $g_HFILE
echo "/* CHAR functions */" >> $g_HFILE
echo "#include <ctype.h>" >> $g_HFILE
echo "char* __b2c__ucase(char *__b2c__s){int __b2c__i; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], strlen(__b2c__s)*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, strlen(__b2c__s)*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "for(__b2c__i=0; __b2c__i < strlen(__b2c__s); __b2c__i++){__b2c__sbuffer[__b2c__sbuffer_ptr][__b2c__i]=toupper(__b2c__s[__b2c__i]);}" >> $g_HFILE
echo "return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define UCASE\$(x) ((x != NULL) ? __b2c__ucase(x) : \"null\")" >> $g_HFILE
echo "char* __b2c__lcase(char *__b2c__s){int __b2c__i; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], strlen(__b2c__s)*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, strlen(__b2c__s)*sizeof(char)+1*sizeof(char));" >> $g_HFILE
echo "for(__b2c__i=0; __b2c__i < strlen(__b2c__s); __b2c__i++){__b2c__sbuffer[__b2c__sbuffer_ptr][__b2c__i]=tolower(__b2c__s[__b2c__i]);}" >> $g_HFILE
echo "return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define LCASE\$(x) ((x != NULL) ? __b2c__lcase(x) : \"null\")" >> $g_HFILE
echo >> $g_HFILE
echo "/* I/O functions */" >> $g_HFILE
echo "char* __b2c__dec2hex(int __b2c__i){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS);" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, $g_MAX_DIGITS);" >> $g_HFILE
echo "snprintf(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%X\", __b2c__i); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "unsigned int __b2c__hex2dec(char *__b2c__h){unsigned int __b2c__i; if(sscanf(__b2c__h, \"%X\", &__b2c__i) == EOF && !__b2c__trap){ERROR = 5;" >> $g_HFILE
echo "fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} return(int)(__b2c__i);}" >> $g_HFILE
echo "unsigned char __b2c__char2asc(char *__b2c__c){return (unsigned char)*__b2c__c;}" >> $g_HFILE
echo "char* __b2c__asc2char(int __b2c__i){__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], 2*sizeof(char));" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, 2*sizeof(char)); sprintf(__b2c__sbuffer[__b2c__sbuffer_ptr], \"%c\", __b2c__i);" >> $g_HFILE
echo "return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "long __b2c__filelen(const char *__b2c__x) {struct stat __b2c__buf; if(__b2c__x == NULL) return 0;" >> $g_HFILE
echo "if(stat(__b2c__x, &__b2c__buf) < 0 && !__b2c__trap){ERROR = 24; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} return(long)(__b2c__buf.st_size);}" >> $g_HFILE
echo "#define FILELEN(x) __b2c__filelen(x)" >> $g_HFILE
echo "long __b2c__filetype(const char *__b2c__x) {struct stat __b2c__buf; if(__b2c__x == NULL) return 0; if(stat(__b2c__x, &__b2c__buf) < 0 && !__b2c__trap)" >> $g_HFILE
echo "{ERROR = 24; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);}" >> $g_HFILE
echo "if (S_ISREG(__b2c__buf.st_mode)) return 1; if (S_ISDIR(__b2c__buf.st_mode)) return 2;" >> $g_HFILE
echo "if (S_ISCHR(__b2c__buf.st_mode)) return 3; if (S_ISBLK(__b2c__buf.st_mode)) return 4; if (S_ISFIFO(__b2c__buf.st_mode)) return 5;" >> $g_HFILE
echo "if (S_ISLNK(__b2c__buf.st_mode)) return 6; if (S_ISSOCK(__b2c__buf.st_mode)) return 7; return 0;}" >> $g_HFILE
echo "#define FILETYPE(x) __b2c__filetype(x)" >> $g_HFILE
echo "long __b2c__search(FILE* __b2c__x, char* __b2c__y){long __b2c__off; long __b2c__pos; char* __b2c__ptr;" >> $g_HFILE
echo "if(__b2c__x == NULL && !__b2c__trap){ERROR = 2; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);}" >> $g_HFILE
echo "if(__b2c__y == NULL && !__b2c__trap){ERROR = 25; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);}" >> $g_HFILE
echo "__b2c__pos = ftell(__b2c__x); __b2c__ptr = (char*)malloc((strlen(__b2c__y)+1)*sizeof(char)); __b2c__off = 0; do {" >> $g_HFILE
echo "fseek(__b2c__x, __b2c__off, SEEK_SET);memset(__b2c__ptr, 0, sizeof(char)*(strlen(__b2c__y)+1));" >> $g_HFILE
echo "fread(__b2c__ptr, sizeof(char), strlen(__b2c__y), __b2c__x);__b2c__off+=1;" >> $g_HFILE
echo "} while(!feof(__b2c__x) && strncmp(__b2c__ptr, __b2c__y, strlen(__b2c__y)));" >> $g_HFILE
echo "if(strncmp(__b2c__ptr, __b2c__y, strlen(__b2c__y))) __b2c__off = 0; fseek(__b2c__x, __b2c__pos, SEEK_SET); free(__b2c__ptr); return (--__b2c__off);}" >> $g_HFILE
echo "#define SEARCH(x, y) __b2c__search(x, y)" >> $g_HFILE
echo "#define ENDFILE(x) feof(x)" >> $g_HFILE
echo "#define TELL(x) ftell(x)" >> $g_HFILE
echo "#define HEX\$(x) __b2c__dec2hex(x)" >> $g_HFILE
echo "#define DEC(x) __b2c__hex2dec(x)" >> $g_HFILE
echo "#define ASC(x) __b2c__char2asc(x)" >> $g_HFILE
echo "#define CHR\$(__b2c__x) __b2c__asc2char(__b2c__x)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Dynamic loading, errors */" >> $g_HFILE
echo "#include <dlfcn.h>" >> $g_HFILE
echo "#include <errno.h>" >> $g_HFILE
echo >> $g_HFILE
echo "/* GETKEY */" >> $g_HFILE
echo "#include <termios.h>" >> $g_HFILE
echo "long __b2c__getch(){long __b2c__ch; struct termios __b2c__oldt, __b2c__newt; tcgetattr(0, &__b2c__oldt);" >> $g_HFILE
echo "__b2c__newt = __b2c__oldt; __b2c__newt.c_lflag &= ~(ICANON | ECHO); tcsetattr(0, TCSANOW, &__b2c__newt);" >> $g_HFILE
echo "__b2c__ch = getchar(); tcsetattr(0, TCSANOW, &__b2c__oldt); return __b2c__ch;} " >> $g_HFILE
echo "#define GETKEY __b2c__getch()" >> $g_HFILE
echo >> $g_HFILE
echo "/* Constants, logical stuff */" >> $g_HFILE
echo "#define PI 3.14159265" >> $g_HFILE
echo "#define NOT(x) (!(x))" >> $g_HFILE
echo "#define AND &&" >> $g_HFILE
echo "#define OR ||" >> $g_HFILE
echo "#define EQ ==" >> $g_HFILE
echo "#define NE !=" >> $g_HFILE
echo "#define TRUE 1" >> $g_HFILE
echo "#define FALSE 0" >> $g_HFILE
echo "#define NL\$ \"\n\"" >> $g_HFILE
echo "#define STRING char*" >> $g_HFILE
echo "#define NUMBER long" >> $g_HFILE
echo "#define FLOATING double" >> $g_HFILE
echo "#define ISTRUE(x) (x!=0)" >> $g_HFILE
echo "#define ISFALSE(x) (x==0)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Date and time */" >> $g_HFILE
echo "#include <time.h>" >> $g_HFILE
echo "#define NOW (long)time(NULL)" >> $g_HFILE
if [[ `uname` = "OSF1" ]]
then
    echo "long __b2c__time(int __b2c__now, int __b2c__which){struct tm *ts; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
else
    echo "long __b2c__time(long __b2c__now, int __b2c__which){struct tm *ts; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
fi
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS);" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, $g_MAX_DIGITS); ts = localtime(&__b2c__now); switch(__b2c__which) { case 1: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%d\", ts); break;" >> $g_HFILE
echo "case 2: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%m\", ts); break; case 3: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%Y\", ts); break;" >> $g_HFILE
echo "case 4: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%H\", ts); break; case 5: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%M\", ts); break;" >> $g_HFILE
echo "case 6: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%S\", ts); break; case 7: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%W\", ts); break;}" >> $g_HFILE
echo "return(atol(__b2c__sbuffer[__b2c__sbuffer_ptr]));}" >> $g_HFILE
echo "#define DAY(x) __b2c__time(x, 1)" >> $g_HFILE
echo "#define MONTH(x) __b2c__time(x, 2)" >> $g_HFILE
echo "#define YEAR(x) __b2c__time(x, 3)" >> $g_HFILE
echo "#define HOUR(x) __b2c__time(x, 4)" >> $g_HFILE
echo "#define MINUTE(x) __b2c__time(x, 5)" >> $g_HFILE
echo "#define SECOND(x) __b2c__time(x, 6)" >> $g_HFILE
echo "#define WEEK(x) __b2c__time(x, 7)" >> $g_HFILE
if [[ `uname` = "OSF1" ]]
then
    echo "char * __b2c__datename(int __b2c__now, int __b2c__which){struct tm *ts; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
else
    echo "char * __b2c__datename(long __b2c__now, int __b2c__which){struct tm *ts; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
fi
echo "__b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS);" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, $g_MAX_DIGITS); ts = localtime(&__b2c__now); switch(__b2c__which){case 1: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%A\", ts); break;" >> $g_HFILE
echo "case 2: strftime(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_MAX_DIGITS, \"%B\", ts); break;} return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define WEEKDAY\$(x) __b2c__datename(x, 1)" >> $g_HFILE
echo "#define MONTH\$(x) __b2c__datename(x, 2)" >> $g_HFILE
echo "unsigned long __b2c__epoch(int __b2c__year, int __b2c__month, int __b2c__day, int __b2c__hour, int __b2c__minute, int __b2c__second){" >> $g_HFILE
echo "struct tm tm; time_t __b2c__t; tm.tm_year = __b2c__year - 1900; tm.tm_mon = __b2c__month - 1; tm.tm_mday = __b2c__day;" >> $g_HFILE
echo "tm.tm_hour = __b2c__hour; tm.tm_min = __b2c__minute; tm.tm_sec = __b2c__second;" >> $g_HFILE
echo "tm.tm_isdst = -1; __b2c__t = mktime(&tm); if (__b2c__t == -1) return (0); return(long) __b2c__t; }" >> $g_HFILE
echo "#define TIMEVALUE(x,y,z,a,b,c) __b2c__epoch(x,y,z,a,b,c)" >> $g_HFILE
echo "char* __b2c__os(void){struct utsname __b2c__buf; __b2c__sbuffer_ptr++; if(__b2c__sbuffer_ptr >= $g_MAX_BUFFERS) __b2c__sbuffer_ptr=0;" >> $g_HFILE
echo "__b2c__sbuffer[__b2c__sbuffer_ptr] = (char*)realloc(__b2c__sbuffer[__b2c__sbuffer_ptr], $g_BUFFER_SIZE);" >> $g_HFILE
echo "memset(__b2c__sbuffer[__b2c__sbuffer_ptr], 0, $g_BUFFER_SIZE); if(uname(&__b2c__buf) < 0 && !__b2c__trap) {ERROR = 26; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);}" >> $g_HFILE
echo "strcpy(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__buf.sysname); strcat(__b2c__sbuffer[__b2c__sbuffer_ptr], \" \");" >> $g_HFILE
echo "strcat(__b2c__sbuffer[__b2c__sbuffer_ptr], __b2c__buf.machine); return(__b2c__sbuffer[__b2c__sbuffer_ptr]);}" >> $g_HFILE
echo "#define OS\$ __b2c__os()" >> $g_HFILE
echo >> $g_HFILE
echo "/* Peek and Poke */" >> $g_HFILE
echo "struct stat __b2c__sb;" >>  $g_HFILE
echo "void __b2c__memory__check(char* __b2c__x) {if (stat(__b2c__x, &__b2c__sb) == -1 && errno == EFAULT) {ERROR = 1;} }" >> $g_HFILE
echo "long __b2c__malloc(long __b2c__x) {void *__b2c__mem; if(__b2c__x==0) return(0); __b2c__mem = malloc(sizeof(char)*(__b2c__x));" >> $g_HFILE
echo "if(__b2c__mem == NULL && !__b2c__trap) {ERROR = 6; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} memset(__b2c__mem, 0, sizeof(char)*(__b2c__x));return(long)__b2c__mem;}" >> $g_HFILE
echo "#define MEMORY(__b2c__x) __b2c__malloc(__b2c__x)" >> $g_HFILE
echo "void __b2c__realloc(void* __b2c__x, long __b2c__y) {if(__b2c__x == NULL) return; if(!__b2c__trap) {__b2c__memory__check((char*)__b2c__x);" >> $g_HFILE
echo "if(ERROR) {fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} } else {__b2c__x = realloc((void*)__b2c__x, sizeof(char)*(__b2c__y));" >> $g_HFILE
echo "if(__b2c__x == NULL && !__b2c__trap) {ERROR = 6; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} return;} }" >> $g_HFILE
echo "#define RESIZE(__b2c__x, __b2c__y) __b2c__realloc((void*)__b2c__x, __b2c__y)" >> $g_HFILE
echo "void __b2c__poke(char* __b2c__x, unsigned char __b2c__y) {if (!__b2c__trap){__b2c__memory__check((char*)__b2c__x); if(ERROR) {" >> $g_HFILE
echo "fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} } *__b2c__x = (unsigned char)__b2c__y;}" >> $g_HFILE
echo "#define POKE(__b2c__x, __b2c__y) __b2c__poke((char*)(__b2c__x), (unsigned char)(__b2c__y))" >> $g_HFILE
echo "unsigned char __b2c__peek(char* __b2c__x) {if(!__b2c__trap) {__b2c__memory__check((char*) __b2c__x); if(ERROR) {" >> $g_HFILE
echo "fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} } return((unsigned char)*__b2c__x);}" >> $g_HFILE
echo "#define PEEK(__b2c__x) __b2c__peek((char*)(__b2c__x))" >> $g_HFILE
echo "#define ADDRESS(__b2c__x) (long)(&__b2c__x)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Network Select */" >> $g_HFILE
echo "long __b2c__netpeek(int __b2c__fd, int __b2c__usec){fd_set __b2c__rfds;struct timeval __b2c__tv;long __b2c__retval;" >> $g_HFILE
echo "FD_ZERO(&__b2c__rfds); FD_SET(__b2c__fd, &__b2c__rfds);__b2c__tv.tv_usec = __b2c__usec*1000;__b2c__tv.tv_sec = 0;" >> $g_HFILE
echo "__b2c__retval = select(__b2c__fd + 1, &__b2c__rfds, NULL, NULL, &__b2c__tv);" >> $g_HFILE
echo "if (__b2c__retval == -1 && !__b2c__trap) {ERROR = 16; fprintf(stderr, \"%s\n\", ERR\$(ERROR)); exit(ERROR);} return(__b2c__retval);}" >> $g_HFILE
echo "#define WAIT(x, y) __b2c__netpeek(x, y)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Regex */" >> $g_HFILE
echo "#include <regex.h>" >> $g_HFILE
echo "long __b2c__regex(char* __b2c__x, char* __b2c__y){regex_t __b2c__reg; int __b2c__reti; char __b2c__buf[100]; __b2c__reti = regcomp(&__b2c__reg, __b2c__y, REG_EXTENDED|REG_NOSUB);" >> $g_HFILE
echo "if(!__b2c__trap && __b2c__reti){ERROR=27; fprintf(stderr, \"%s: \", ERR\$(ERROR)); regerror(__b2c__reti, &__b2c__reg, __b2c__buf, sizeof(__b2c__buf)); fprintf(stderr, \"%s\n\", __b2c__buf);" >> $g_HFILE
echo "exit(ERROR);} __b2c__reti = regexec(&__b2c__reg, __b2c__x, 0, NULL, 0);" >> $g_HFILE
echo "regfree(&__b2c__reg); if( !__b2c__reti ) return (1); else return (0);}" >> $g_HFILE
echo "#define REGEX(x, y) __b2c__regex(x, y)" >> $g_HFILE
echo >> $g_HFILE
echo "/* Declare reserved variable 'ARGUMENT\$' */" >> $g_HFILE
echo "int __b2c__counter;" >> $g_HFILE
echo "int __b2c__arglen = 0;" >> $g_HFILE
echo "char *ARGUMENT\$;" >> $g_HFILE
echo >> $g_HFILE
echo "/* Initialize stack arrays and pointer */" >> $g_HFILE
echo "char *__b2c__stringstack[$g_STACK_DEPTH];" >> $g_HFILE
echo "double __b2c__doublestack[$g_STACK_DEPTH];" >> $g_HFILE
echo "long __b2c__longstack[$g_STACK_DEPTH];" >> $g_HFILE
echo "int __b2c__typestack[$g_STACK_DEPTH];" >> $g_HFILE
echo "int __b2c__stackptr;" >> $g_HFILE
# Signal function
echo "/* Signal trapping activated with TRAP */" >> $g_HFILE
echo "void __b2c__catch_signal(int sig){" >> $g_HFILE
echo "switch (sig) {case SIGABRT: fprintf(stderr, \"ERROR: signal ABORT received - internal error. Try to compile the program with TRAP LOCAL to find the cause.\n\"); break;" >> $g_HFILE
echo "case SIGFPE: fprintf(stderr, \"ERROR: signal for FPE received - division by zero? Examine the calculations in the program.\n\"); break;" >> $g_HFILE
echo "case SIGSEGV: fprintf(stderr, \"ERROR: signal for SEGMENTATION FAULT received - memory invalid or array out of bounds? Try to compile the program with TRAP LOCAL to find the cause.\n\"); break;" >> $g_HFILE
echo "case SIGILL: fprintf(stderr, \"ERROR: signal for ILLEGAL INSTRUCTION received - executing the program on other hardware? Try to recompile the program from scratch.\n\"); break;} exit(sig);}" >> $g_HFILE
# Add error function
echo >> $g_HFILE
echo "/* Initialize error function */" >> $g_HFILE
echo "char *ERR\$(int __b2c__nr){static char __b2c__warn[$g_BUFFER_SIZE];" >> $g_HFILE
echo "switch(__b2c__nr){" >> $g_HFILE
echo "case 0: strcpy(__b2c__warn,\"Success\"); break;" >> $g_HFILE
echo "case 1: strcpy(__b2c__warn,\"Trying to access illegal memory: \"); strncat(__b2c__warn,strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 2: strcpy(__b2c__warn,\"Error opening file: \"); strncat(__b2c__warn,strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 3: strcpy(__b2c__warn, \"Could not open library \"); strncat(__b2c__warn, dlerror(), $g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 4: strcpy(__b2c__warn, \"Symbol not found in library \"); strncat(__b2c__warn, dlerror(), $g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 5: strcpy(__b2c__warn, \"Wrong hexvalue: \"); strncat(__b2c__warn, strerror(errno), $g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 6: strcpy(__b2c__warn, \"Unable to claim memory.\"); break;" >> $g_HFILE
echo "case 7: strcpy(__b2c__warn, \"Unable to delete file: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 8: strcpy(__b2c__warn, \"Could not open directory: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 9: strcpy(__b2c__warn, \"Unable to rename file: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 10: strcpy(__b2c__warn, \"NETWORK argument should contain colon with port number\"); break;" >> $g_HFILE
echo "case 11: strcpy(__b2c__warn, \"Could not resolve hostname: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 12: strcpy(__b2c__warn, \"Socket error: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 13: strcpy(__b2c__warn, \"Unable to open address: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 14: strcpy(__b2c__warn, \"Error reading from socket: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 15: strcpy(__b2c__warn, \"Error sending to socket: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 16: strcpy(__b2c__warn, \"Error checking socket: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 17: strcpy(__b2c__warn, \"Unable to bind the specified socket address: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 18: strcpy(__b2c__warn, \"Unable to listen to socket address: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 19: strcpy(__b2c__warn, \"Cannot accept incoming connection: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 20: strcpy(__b2c__warn, \"Unable to remove directory: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 21: strcpy(__b2c__warn, \"Unable to create directory: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 22: strcpy(__b2c__warn, \"Unable to change to directory: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 23: strcpy(__b2c__warn, \"GETENVIRON argument does not exist as environment variable\"); break;" >> $g_HFILE
echo "case 24: strcpy(__b2c__warn, \"Unable to stat file: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 25: strcpy(__b2c__warn, \"Search contains illegal string\"); break;" >> $g_HFILE
echo "case 26: strcpy(__b2c__warn, \"Cannot return OS name: \"); strncat(__b2c__warn, strerror(errno),$g_BUFFER_SIZE-48); break;" >> $g_HFILE
echo "case 27: strcpy(__b2c__warn, \"Illegal regex expression\"); break;" >> $g_HFILE
echo "}; ERROR = 0; return(__b2c__warn);}" >> $g_HFILE
echo >> $g_HFILE
echo "/* User program definitions */" >> $g_HFILE

# Initialize the arrayfiles for DATA statement
echo "char *__b2c__stringarray[] = {" > $STRINGARRAYFILE
echo "int __b2c__intarray[] = {" > $INTARRAYFILE
echo "double __b2c__floatarray[] = {" > $FLOATARRAYFILE

# There are no imported symbols yet
g_IMPORTED=

# This is the current file we are parsing
g_CURFILE=$SOURCEFILE

# Initialize
LEN=
TOTAL=
SEQ=
g_COUNTER=1

# Start walking through program
while read -r LINE
do
    echo -e -n "\rStarting conversion... $g_COUNTER"

    # Line is not empty?
    if [[ -n "$LINE" ]]
    then
	if [[ "$LINE" = +(* \\) && "$LINE" != +(REM*) && "$LINE" != +(\'*) ]]
	then
	    let LEN="${#LINE}"-2
	    SEQ="${LINE:0:$LEN}"
	    TOTAL=$TOTAL$SEQ
	else
	    Parse_Line "$TOTAL$LINE"
	    TOTAL=
	fi
    fi
    ((g_COUNTER+=1))

done < $SOURCEFILE

# Finalize main C-file
echo "__B2C__PROGRAM__EXIT:" >> $g_CFILE
echo "return 0;" >> $g_CFILE
echo "}" >> $g_CFILE

# Finalize INT ARRAY file for DATA
echo " \"\" };" >> $STRINGARRAYFILE

# Finalize STRING ARRAY file for DATA
echo " 0};" >> $INTARRAYFILE

# Finalize FLOAT ARRAY file for DATA
echo " 0.0};" >> $FLOATARRAYFILE

# Include functions and subs
echo -e "$g_INCLUDE_FILES" >> $g_HFILE

echo -e "\rStarting conversion... done.  "

# Start indentation
if [[ -n `which indent 2>/dev/null` ]]
then
    echo -n "Applying indentation... "
    for i in $g_TMP_FILES
    do
	indent $i
	if [[ `uname` = "Darwin" || `uname` = +(*BSD*) ]]
	then
	    rm `basename $i`.BAK
	else
	    rm $i~
	fi
    done
    echo "done."
else
    echo "WARNING: 'indent' not found on this system!"
    echo "Generated source code cannot be beautified."
fi

# Start compilation
if [[ $g_NO_COMPILE -eq 0 ]]
then
    echo -n "Starting compilation... "
    $g_CCNAME $g_CCFLAGS -o ${SOURCEFILE%.*} $g_CFILE $g_LDFLAGS > $g_TEMPDIR/${SOURCEFILE}.log 2>&1

    g_TMP_FILES="$g_TMP_FILES $g_TEMPDIR/${SOURCEFILE}.log"

    if [[ -z `cat $g_TEMPDIR/${SOURCEFILE}.log` ]]
    then
	echo "done."
	echo "Program '${SOURCEFILE%.*}' ready."
    else
	# Only print first error
	echo -e "Compiler emits messages!"
	while read -r g_ERROR
	do
	    if [[ $g_ERROR != +(---*) && -n `echo "${g_ERROR}" | grep -E "\:[0-9]+\: error\:"` ]]
	    then
		echo -e "\n${g_ERROR}"
		break
	    fi
	done < $g_TEMPDIR/${SOURCEFILE}.log

	# Non-gcc
	if [[ -z $g_ERROR || $g_CCFLAGS = +(*Wall*) ]]
	then
	    echo
	    cat $g_TEMPDIR/${SOURCEFILE}.log
	    echo
	else
	    # Get the error string from the logfile
	    g_FILE_LINE=${g_ERROR%%: *}
	    g_COUNTER=1
	    while read -r LINE
	    do
		if [[ $g_COUNTER -eq ${g_FILE_LINE##*:} ]]
		then
		    echo -e " ---> ${LINE} <---\n"
		    break
		fi
		((g_COUNTER+=1))
	    done < ${g_FILE_LINE%%:*}
	fi
    fi
fi

# Cleanup
if [[ $g_TMP_PRESERVE -eq 0 ]]
then
    for i in $g_TMP_FILES
    do
	rm $i
    done
fi

exit 0
